"""Classes in this module are used to declare the place where a xml value is located inside a document.

They also provide a mapping between XML data types (which are always stings in specific formats) and
python types. By doing so these classes completely hide the XML nature of data.
The basic offered types are Element, list of elements, attribute, and list of attributes.
They are the buildings blocks that are needed to declare XML data types.
Container properties represent values in xml nodes.
"""

from __future__ import annotations

import copy
import time
from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import TYPE_CHECKING, Any, Callable

from lxml import etree

from sdc11073 import xml_utils
from sdc11073.exceptions import ApiUsageError
from sdc11073.namespaces import QN_TYPE, docname_from_qname, text_to_qname

from . import isoduration
from .dataconverters import (
    BooleanConverter,
    ClassCheckConverter,
    DecimalConverter,
    DurationConverter,
    EnumConverter,
    IntegerConverter,
    ListConverter,
    NullConverter,
    StringConverter,
    TimestampConverter,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from decimal import Decimal

    from sdc11073.mdib.containerbase import ContainerBase
    from sdc11073.namespaces import NamespaceHelper
    from sdc11073.xml_types.basetypes import XMLTypeBase

    from .dataconverters import DataConverterProtocol
    from .isoduration import DurationType

STRICT_TYPES = True  # if True, only the expected types are excepted.
MANDATORY_VALUE_CHECKING = True  # checks if mandatory values are present when xml is generated


class ElementNotFoundError(Exception):  # noqa: D101
    pass


class _NumberStack:
    # uses as a part of _local_var_name in _XmlStructureBaseProperty.
    # This makes duplicate names impossible
    _value = 0

    @classmethod
    def unique_number(cls) -> str:
        cls._value += 1
        return str(cls._value)


class _XmlStructureBaseProperty(ABC):
    """_XmlStructureBaseProperty defines a python property that converts between Python Data Types and XML data types.

    It has knowledge about two things:
    - how to covert data from xml to python type and vice versa
    - name/ location of the xml data in a node.

    All derived Properties have the same interface:
    __get__ and __set__ : read and write access, using Python data types.
    get_py_value_from_node: reads the value from XML data and converts it to Python data type.
    update_xml_value: convert the Python data type to XML type and write it to XML node.
    """

    def __init__(
        self,
        local_var_name: str,
        value_converter: DataConverterProtocol,
        default_py_value: Any | None = None,
        implied_py_value: Any | None = None,
        is_optional: bool = False,
    ):
        """Construct an instance.

        :param local_var_name: a member with this same is added to instance
        :param value_converter: DataConverterProtocol
        :param default_py_value: initial value when initialized
                                 (should be set for mandatory elements, otherwise created xml might violate schema)
                                 and if the xml element does not exist.
        :param implied_py_value: for optional elements, this is the value that shall be implied if
                                 xml element does not exist.
                                 This value is for information only! Access only via class possible.
        :param is_optional: reflects if this element is optional in schema
        """
        if implied_py_value is not None and default_py_value is not None:
            raise ValueError('set only one of default_py_value and implied_py_value')
        if not is_optional and implied_py_value is not None:
            raise ValueError('is_optional == False and implied_py_value != None is not allowed ')
        if not hasattr(value_converter, 'check_valid'):
            raise TypeError
        self._converter = value_converter
        if STRICT_TYPES:
            if default_py_value is not None:
                self._converter.check_valid(default_py_value)
            if implied_py_value is not None:
                self._converter.check_valid(implied_py_value)
        self._default_py_value = None
        self._implied_py_value = None
        if default_py_value is not None:
            self._default_py_value = default_py_value
        if implied_py_value is not None:
            self._implied_py_value = implied_py_value
        self._is_optional = is_optional
        self._local_var_name = local_var_name
        self._is_default_value_set = False

    @property
    def is_optional(self) -> bool:
        return self._is_optional

    def __get__(self, instance, owner) -> Any:  # noqa: ANN001
        """Return a python value, use the locally stored value."""
        if instance is None:  # if called via class
            return self
        try:
            value = getattr(instance, self._local_var_name)
        except AttributeError:
            value = None
        if value is None:
            value = self._implied_py_value
        return value

    def get_actual_value(self, instance: Any) -> Any | None:
        """Return the actual value without considering default value and implied value.

        E.g. return None if no value in xml exists.
        :param instance: the instance that has the property as member
        """
        try:
            return getattr(instance, self._local_var_name)
        except AttributeError:
            return None

    def __set__(self, instance, py_value):  # noqa: ANN001
        """Value is the representation on the program side, e.g a float."""
        if STRICT_TYPES:
            self._converter.check_valid(py_value)
        setattr(instance, self._local_var_name, py_value)

    def init_instance_data(self, instance: Any):
        """Set initial values to default_py_value.

        This method is used internally and should not be called by application.
        :param instance: the instance that has the property as member
        :return: None
        """
        if self._default_py_value is not None:
            setattr(instance, self._local_var_name, copy.deepcopy(self._default_py_value))

    @abstractmethod
    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Update node with current data from instance.

        This method is used internally and should not be called by application.
        :param instance: the instance that has the property as member
        :param node: the etree node that shall be updated
        :return: None
        """

    @abstractmethod
    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement):
        """Read data from node.

        This method is used internally and should not be called by application.
        :param instance: the instance that has the property as member
        :param node: the etree node that provides the value
        :return: value
        """

    def update_from_node(self, instance: Any, node: xml_utils.LxmlElement):
        """Update instance data with data from node.

        This method is used internally and should not be called by application.
        :param instance:the instance that has the property as member
        :param node:the etree node that provides the value
        :return: value
        :return:
        """
        value = self.get_py_value_from_node(instance, node)
        setattr(instance, self._local_var_name, value)


class _AttributeBase(_XmlStructureBaseProperty):
    """Base class that represents an XML Attribute.

    The XML Representation is a string.
    The python representation is determined by value_converter.
    """

    def __init__(
        self,
        attribute_name: str,
        value_converter: DataConverterProtocol | None = None,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = True,
    ):
        """Construct an instance.

        :param attribute_name: name of the attribute in xml node
        :param value_converter: converter between xml value and python value
        :param default_py_value: see base class doc.
        :param implied_py_value: see base class doc.
        :param is_optional: see base class doc.
        """
        if isinstance(attribute_name, etree.QName):
            local_var_name = f'_a_{attribute_name.localname}_{_NumberStack.unique_number()}'
        else:
            local_var_name = f'_a_{attribute_name.lower()}_{_NumberStack.unique_number()}'
        super().__init__(local_var_name, value_converter, default_py_value, implied_py_value, is_optional)
        self._attribute_name = attribute_name

    def get_py_value_from_node(
        self,
        instance: Any,  # noqa: ARG002
        node: xml_utils.LxmlElement | None,
    ) -> Any:
        xml_value = None if node is None else node.attrib.get(self._attribute_name)
        return None if xml_value is None else self._converter.to_py(xml_value)

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            # this can only happen if there is no default value defined and __set__ has never been called
            py_value = None
        if py_value is None:
            if MANDATORY_VALUE_CHECKING and not self.is_optional:
                raise ValueError(f'mandatory value {self._attribute_name} missing')  # noqa: EM102
            try:
                if self._attribute_name in node.attrib:
                    del node.attrib[self._attribute_name]
            except ElementNotFoundError:
                return
        else:
            xml_value = self._converter.to_xml(py_value)
            node.set(self._attribute_name, xml_value)

    def __str__(self) -> str:
        return f'{self.__class__.__name__} attribute {self._attribute_name}'


class _ElementBase(_XmlStructureBaseProperty, ABC):
    """_ElementBase represents an XML Element."""

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        value_converter: DataConverterProtocol,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = False,
    ):
        """Construct the representation of a (sub) element in xml.

        :param sub_element_name: a QName or None. If None, the property represents the node itself,
                                 otherwise the sub node with given name.
        :param value_converter: see base class doc.
        :param default_py_value: see base class doc.
        :param implied_py_value: see base class doc.
        :param is_optional: see base class doc.
        """
        if sub_element_name is None:
            local_var_name = f'_e_{_NumberStack.unique_number()}'
        else:
            local_var_name = f'_e_{sub_element_name.localname.lower()}_{_NumberStack.unique_number()}'
        super().__init__(local_var_name, value_converter, default_py_value, implied_py_value, is_optional)
        self._sub_element_name = sub_element_name

    @staticmethod
    def _get_element_by_child_name(
        node: xml_utils.LxmlElement,
        sub_element_name: etree.QName | None,
        create_missing_nodes: bool,
    ) -> xml_utils.LxmlElement:
        if sub_element_name is None:
            return node
        sub_node = node.find(sub_element_name)
        if sub_node is None:
            if not create_missing_nodes:
                raise ElementNotFoundError(f'Element {sub_element_name} not found in {node.tag}')  # noqa: EM102
            sub_node = etree.SubElement(node, sub_element_name)  # create this node
        return sub_node

    def remove_sub_element(self, node: xml_utils.LxmlElement):
        if self._sub_element_name is None:
            return
        sub_node = node.find(self._sub_element_name)
        if sub_node is not None:
            node.remove(sub_node)

    def __str__(self) -> str:
        return f'{self.__class__.__name__} in sub element {self._sub_element_name}'


class StringAttributeProperty(_AttributeBase):
    """Python representation is a string."""

    def __init__(
        self,
        attribute_name: str,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = True,
    ):
        super().__init__(attribute_name, StringConverter, default_py_value, implied_py_value, is_optional)


class AnyURIAttributeProperty(StringAttributeProperty):
    """Represents an AnyURIAttribute."""


class CodeIdentifierAttributeProperty(StringAttributeProperty):
    """Represents a CodeIdentifier attribute."""


class HandleAttributeProperty(StringAttributeProperty):
    """Represents a Handle attribute."""


class HandleRefAttributeProperty(StringAttributeProperty):
    """Represents a HandleRef attribute."""


class SymbolicCodeNameAttributeProperty(StringAttributeProperty):
    """Represents a SymbolicCodeName attribute."""


class ExtensionAttributeProperty(StringAttributeProperty):
    """Represents an Extension attribute."""


class LocalizedTextRefAttributeProperty(StringAttributeProperty):
    """Represents a LocalizedTextRef attribute."""


class TimeZoneAttributeProperty(StringAttributeProperty):
    """Represents a TimeZone attribute."""


class EnumAttributeProperty(_AttributeBase):
    """Base class for enum attributes."""

    def __init__(
        self,
        attribute_name: str,
        enum_cls: Any,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = True,
    ):
        super().__init__(attribute_name, EnumConverter(enum_cls), default_py_value, implied_py_value, is_optional)


class TimestampAttributeProperty(_AttributeBase):
    """Represents a Timestamp attribute.

    XML notation is integer in milliseconds.
    Python is a float in seconds.
    """

    def __init__(
        self,
        attribute_name: str,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=TimestampConverter,
            default_py_value=default_py_value,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )


class CurrentTimestampAttributeProperty(_AttributeBase):
    """Represents a special Timestamp attribute used for ClockState, it always writes current time to node.

    Setting the value from python is possible, but makes no sense.
    """

    def __init__(self, attribute_name: str, is_optional: bool = True):
        super().__init__(
            attribute_name,
            value_converter=TimestampConverter,
            default_py_value=None,
            is_optional=is_optional,
        )

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        setattr(instance, self._local_var_name, time.time())
        super().update_xml_value(instance, node)


class DecimalAttributeProperty(_AttributeBase):
    """Represents a Decimal attribute."""

    def __init__(
        self,
        attribute_name: str,
        default_py_value: Decimal | None = None,
        implied_py_value: Decimal | None = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=DecimalConverter,
            default_py_value=default_py_value,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )


class QualityIndicatorAttributeProperty(DecimalAttributeProperty):
    """Represents a QualityIndicator attribute, a value between 0 and 1."""


class DurationAttributeProperty(_AttributeBase):
    """Represents a Duration attribute.

    XML notation is integer in milliseconds.
    Python is a float in seconds.
    """

    def __init__(
        self,
        attribute_name: str,
        default_py_value: DurationType | None = None,
        implied_py_value: DurationType | None = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=DurationConverter,
            default_py_value=default_py_value,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )


class IntegerAttributeProperty(_AttributeBase):
    """Represents an Integer attribute.

    XML notation is an integer, python is an integer.
    """

    def __init__(
        self,
        attribute_name: str,
        default_py_value: int | None = None,
        implied_py_value: int | None = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=IntegerConverter,
            default_py_value=default_py_value,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )


class UnsignedIntAttributeProperty(IntegerAttributeProperty):
    """Represents an UnsignedInt attribute.

    Python has no unsigned int, therefore this is the same as IntegerAttributeProperty.
    """


class VersionCounterAttributeProperty(UnsignedIntAttributeProperty):
    """Represents a VersionCounter attribute.

    VersionCounter in BICEPS is unsigned long.
    Python has no unsigned long, therefore this is the same as IntegerAttributeProperty.
    """


class ReferencedVersionAttributeProperty(VersionCounterAttributeProperty):
    """Represents an ReferencedVersion attribute."""


class BooleanAttributeProperty(_AttributeBase):
    """Represents a Boolean attribute.

    XML notation is 'true' or 'false'.
    Python is a bool.
    """

    def __init__(
        self,
        attribute_name: str,
        default_py_value: bool | None = None,
        implied_py_value: bool | None = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=BooleanConverter,
            default_py_value=default_py_value,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )


class QNameAttributeProperty(_AttributeBase):
    """Represents a qualified name attribute.

    XML Representation is a prefix:name string, Python representation is a QName.
    """

    def __init__(
        self,
        attribute_name: str,
        implied_py_value: etree.QName | None = None,
        is_optional: bool = True,
    ):
        super().__init__(
            attribute_name,
            value_converter=ClassCheckConverter(etree.QName),
            default_py_value=None,
            implied_py_value=implied_py_value,
            is_optional=is_optional,
        )

    def get_py_value_from_node(  # noqa: D102
        self,
        instance: Any,  # noqa: ARG002
        node: xml_utils.LxmlElement | None,
    ) -> xml_utils.QName | None:
        xml_value = None if node is None else node.attrib.get(self._attribute_name)
        return None if xml_value is None else text_to_qname(xml_value, node.nsmap)

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None
            py_value = None
        if py_value is None:
            if MANDATORY_VALUE_CHECKING and not self.is_optional:
                raise ValueError(f'mandatory value {self._attribute_name} missing')  # noqa: EM102
            try:
                if self._attribute_name in node.attrib:
                    del node.attrib[self._attribute_name]
            except ElementNotFoundError:
                return
        else:
            node.set(self._attribute_name, py_value)


class _AttributeListBase(_AttributeBase):
    """Base class for a list of values as attribute.

    XML Representation is a string which is a space separated list.
    Python representation is a list of Any (type depends on ListConverter),
    else a list of converted values.
    """

    _converter: ListConverter

    def __init__(self, attribute_name: str, value_converter: ListConverter, is_optional: bool = True):
        super().__init__(attribute_name, value_converter, is_optional=is_optional)

    def __get__(self, instance, owner):  # noqa: ANN001
        """Return a python value, use the locally stored value."""
        if instance is None:  # if called via class
            return self
        try:
            return getattr(instance, self._local_var_name)
        except AttributeError:
            setattr(instance, self._local_var_name, [])
            return getattr(instance, self._local_var_name)

    def init_instance_data(self, instance: Any):
        setattr(instance, self._local_var_name, [])

    def get_py_value_from_node(
        self,
        instance: Any,  # noqa: ARG002
        node: xml_utils.LxmlElement | None,
    ) -> list[Any]:
        xml_value = None if node is None else node.attrib.get(self._attribute_name)
        if xml_value is not None:
            split_result = xml_value.split(' ')
            return [self._converter.elem_to_py(val) for val in split_result if val]
        return []

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None
        if not py_value and self.is_optional:  # is None:
            if self._attribute_name in node.attrib:
                del node.attrib[self._attribute_name]
        else:
            if py_value is None:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._attribute_name} missing')  # noqa: EM102
                xml_value = ''
            else:
                xml_value = ' '.join([self._converter.elem_to_xml(v) for v in py_value])
            node.set(self._attribute_name, xml_value)


class _StringAttributeListBase(_AttributeListBase):
    """Base class for a list of strings as attribute.

    XML Representation is a string which is a space separated list.
    Python representation is a list of strings.
    """

    def __init__(self, attribute_name: str, value_converter: DataConverterProtocol | None = None):
        converter = value_converter or ListConverter(ClassCheckConverter(str))
        super().__init__(attribute_name, converter)


class HandleRefListAttributeProperty(_StringAttributeListBase):
    """Represents a list of HandleRef attribute."""


class EntryRefListAttributeProperty(_StringAttributeListBase):
    """Represents a list of EntryRef attribute."""


class OperationRefListAttributeProperty(_StringAttributeListBase):
    """Represents a list of OperationRef attribute."""


class AlertConditionRefListAttributeProperty(_StringAttributeListBase):
    """Represents a list of AlertConditionRef attribute."""


class DecimalListAttributeProperty(_AttributeListBase):
    """Represents a list of Decimal attribute.

    XML representation: an attribute string that represents 0...n decimals, separated with spaces.
    Python representation: List of Decimal if attribute is set (can be an empty list!), otherwise None.
    """

    def __init__(self, attribute_name: str):
        super().__init__(attribute_name, ListConverter(DecimalConverter))


class NodeTextProperty(_ElementBase):
    """Represents the text of an XML Element.

    Python representation depends on value converter.
    """

    def __init__(  # noqa: PLR0913
        self,
        sub_element_name: etree.QName | None,
        value_converter: DataConverterProtocol,
        default_py_value: Any | None = None,
        implied_py_value: Any | None = None,
        is_optional: bool = False,
        min_length: int = 0,
    ):
        super().__init__(sub_element_name, value_converter, default_py_value, implied_py_value, is_optional)
        self._min_length = min_length

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node.

        :return: None if the element was not found, else result of converter.
        """
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
        except ElementNotFoundError:
            return None  # element was not found, return None
        return self._converter.to_py(sub_node.text)

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None
        if py_value is None:
            if MANDATORY_VALUE_CHECKING and not self.is_optional and self._min_length:
                raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102

            if not self._sub_element_name:
                # update text of this element
                node.text = None
            elif self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            else:
                sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
                sub_node.text = None
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            sub_node.text = self._converter.to_xml(py_value)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} in sub-element {self._sub_element_name}'


class NodeStringProperty(NodeTextProperty):
    """Represents the text of an XML Element.

    Python representation is a string.
    libxml sets text of element to None, if text in xml is empty. In this case the python value is an empty string.
    if the xml element that should contain the text does not exist, the python value is None.
    """

    def __init__(
        self,
        sub_element_name: etree.QName | None = None,
        default_py_value: str | None = None,
        implied_py_value: str | None = None,
        is_optional: bool = False,
        min_length: int = 0,
    ):
        super().__init__(sub_element_name, StringConverter, default_py_value, implied_py_value, is_optional, min_length)


class AnyUriTextElement(NodeStringProperty):
    """For now the same as NodeStringProperty ,but later it could be handy to add uri type checking."""


# class LocalizedTextContentProperty(NodeStringProperty):
#     pass


class NodeEnumTextProperty(NodeTextProperty):
    """Represents the text of an XML Element.

    Python representation is an enum.
    """

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        enum_cls: Any,
        default_py_value: Any | None = None,
        implied_py_value: Any | None = None,
        is_optional: bool = False,
    ):
        super().__init__(
            sub_element_name,
            EnumConverter(enum_cls),
            default_py_value,
            implied_py_value,
            is_optional,
            min_length=1,
        )
        self.enum_cls = enum_cls


class NodeEnumQNameProperty(NodeTextProperty):
    """Represents a qualified name as text of an XML Element.

    Python representation is an Enum of QName, XML is prefix:localname.
    """

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        enum_cls: Any,
        default_py_value: Any | None = None,
        implied_py_value: Any | None = None,
        is_optional: bool = False,
    ):
        super().__init__(
            sub_element_name,
            EnumConverter(enum_cls),
            default_py_value,
            implied_py_value,
            is_optional,
            min_length=1,
        )
        self.enum_cls = enum_cls

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            prefix, localname = sub_node.text.split(':')
            namespace = node.nsmap[prefix]
            q_name = etree.QName(namespace, localname)
            return self._converter.to_py(q_name)
        except ElementNotFoundError:
            return self._default_py_value

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None
        if py_value is None:
            if MANDATORY_VALUE_CHECKING and not self.is_optional and self._min_length:
                raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102

            if not self._sub_element_name:
                # update text of this element
                node.text = ''
            elif self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            else:
                sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
                sub_node.text = None
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            sub_node.text = self._converter.to_xml(py_value)


class NodeIntProperty(NodeTextProperty):
    """Python representation is an int."""

    def __init__(
        self,
        sub_element_name: etree.QName | None = None,
        default_py_value: int | None = None,
        implied_py_value: int | None = None,
        is_optional: bool = False,
        min_length: int = 0,
    ):
        super().__init__(
            sub_element_name,
            IntegerConverter,
            default_py_value,
            implied_py_value,
            is_optional,
            min_length,
        )


class NodeDecimalProperty(NodeTextProperty):
    """Python representation is an int."""

    def __init__(
        self,
        sub_element_name: etree.QName | None = None,
        default_py_value: Decimal | None = None,
        implied_py_value: Decimal | None = None,
        is_optional: bool = False,
        min_length: int = 0,
    ):
        super().__init__(
            sub_element_name,
            DecimalConverter,
            default_py_value,
            implied_py_value,
            is_optional,
            min_length,
        )


class NodeDurationProperty(NodeTextProperty):
    """Python representation is an int."""

    def __init__(
        self,
        sub_element_name: etree.QName | None = None,
        default_py_value: isoduration.DurationType | None = None,
        implied_py_value: isoduration.DurationType | None = None,
        is_optional: bool = False,
        min_length: int = 0,
    ):
        super().__init__(
            sub_element_name,
            DurationConverter,
            default_py_value,
            implied_py_value,
            is_optional,
            min_length,
        )


class NodeTextQNameProperty(_ElementBase):
    """The handled data is a single qualified name in the text of an element in the form prefix:localname."""

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        is_optional: bool = False,
    ):
        super().__init__(sub_element_name, ClassCheckConverter(etree.QName), None, is_optional=is_optional)

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            xml_value = sub_node.text
            if xml_value is not None:
                return text_to_qname(xml_value, sub_node.nsmap)
        except ElementNotFoundError:
            pass
        return None

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None

        if py_value is None:
            if not self._sub_element_name:
                # update text of this element
                node.text = ''
            elif self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            else:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
                sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
                sub_node.text = None
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            sub_node.text = py_value  # this adds the namesoace to sube_node.nsmap


def _compare_extension(left: xml_utils.LxmlElement, right: xml_utils.LxmlElement) -> bool:
    # SDPi R0019 and R0020 forbid the usage of XML mixed content or xml schema qname type in extensions
    try:
        if left.tag != right.tag:  # compare expanded names
            return False
        if dict(left.attrib) != dict(right.attrib):  # unclear how lxml _Attrib compares
            return False
    except AttributeError:  # right side is not an Element type because expected attributes are missing
        return False

    # ignore comments
    left_children = [child for child in left if not isinstance(child, etree._Comment)]  # noqa: SLF001
    right_children = [child for child in right if not isinstance(child, etree._Comment)]  # noqa: SLF001

    if len(left_children) != len(right_children):  # compare children count
        return False
    if len(left_children) == 0 and len(right_children) == 0:
        if left.text != right.text:  # mixed content is not allowed. only compare text if there are no children
            return False
    return all(map(_compare_extension, left_children, right_children))  # compare children but keep order


class ExtensionLocalValue(list[xml_utils.LxmlElement]):
    """Value of an extension."""

    compare_method: Callable[[xml_utils.LxmlElement, xml_utils.LxmlElement], bool] = _compare_extension
    """may be overwritten by user if a custom comparison behaviour is required"""

    def __eq__(self, other: Sequence) -> bool:
        try:
            if len(self) != len(other):
                return False
        except TypeError:  # len of other cannot be determined
            return False
        return all(self.__class__.compare_method(left, right) for left, right in zip(self, other))

    def __ne__(self, other: ExtensionLocalValue):
        return not self == other


class ExtensionNodeProperty(_ElementBase):
    """Represents an ext:Extension Element that contains 0...n child elements of any kind.

    The python representation is an ExtensionLocalValue with list of elements.
    """

    def __init__(self, sub_element_name: etree.QName | None, default_py_value: Any | None = None):
        super().__init__(sub_element_name, ClassCheckConverter(ExtensionLocalValue), default_py_value, is_optional=True)

    def __set__(self, instance: Any, value: Iterable):
        if not isinstance(value, ExtensionLocalValue):
            value = ExtensionLocalValue(value)
        super().__set__(instance, value)

    def __get__(self, instance, owner):  # noqa: ANN001
        """Return a python value, uses the locally stored value."""
        if instance is None:  # if called via class
            return self
        try:
            value = getattr(instance, self._local_var_name)
        except AttributeError:
            value = None
        if value is None:
            value = ExtensionLocalValue()
            setattr(instance, self._local_var_name, value)
        return value

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        try:
            extension_nodes = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
        except ElementNotFoundError:
            return ExtensionLocalValue()
        return ExtensionLocalValue(extension_nodes[:])

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node.

        The Extension Element is only added if there is at least one element available in local list.
        """
        try:
            extension_local_value = getattr(instance, self._local_var_name)
        except AttributeError:
            return  # nothing to add
        if extension_local_value is None or len(extension_local_value) == 0:
            return  # nothing to add
        sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
        sub_node.extend(xml_utils.copy_node_wo_parent(x) for x in extension_local_value)


class AnyEtreeNodeProperty(_ElementBase):
    """Represents an Element that contains xml tree of any kind."""

    def __init__(self, sub_element_name: etree.QName | None, is_optional: bool = False):
        super().__init__(sub_element_name, NullConverter, default_py_value=None, is_optional=is_optional)

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
        except ElementNotFoundError:
            return None
        return sub_node[:]  # all children

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None

        if py_value is None:
            if self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            elif MANDATORY_VALUE_CHECKING:
                raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            if isinstance(py_value, etree._Element):  # noqa: SLF001
                sub_node.append(py_value)
            else:
                sub_node.extend(py_value)


class SubElementProperty(_ElementBase):
    """Uses a value that has an "as_etree_node" method."""

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        value_class: type[XMLTypeBase],
        default_py_value: Any | None = None,
        implied_py_value: Any | None = None,
        is_optional: bool = False,
    ):
        super().__init__(
            sub_element_name,
            ClassCheckConverter(value_class),
            default_py_value,
            implied_py_value,
            is_optional,
        )
        self.value_class = value_class

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        value = self._default_py_value
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            value_class = self.value_class.value_class_from_node(sub_node)
            value = value_class.from_node(sub_node)
        except ElementNotFoundError:
            pass
        return value

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            py_value = self._default_py_value

        if py_value is None:
            if not self.is_optional:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
                etree.SubElement(node, self._sub_element_name, nsmap=node.nsmap)
        else:
            sub_node = py_value.as_etree_node(self._sub_element_name, node.nsmap, node)
            if (
                hasattr(py_value, 'NODETYPE')
                and hasattr(self.value_class, 'NODETYPE')
                and py_value.NODETYPE != self.value_class.NODETYPE
            ):
                # set xsi type
                sub_node.set(QN_TYPE, docname_from_qname(py_value.NODETYPE, node.nsmap))


class ContainerProperty(_ElementBase):
    """ContainerProperty supports xsi:type information from xml and instantiates value accordingly."""

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        value_class: type[ContainerBase],
        cls_getter: Callable[[etree.QName], type],
        ns_helper: NamespaceHelper,
        is_optional: bool = False,
    ):
        """Construct a ContainerProperty.

        :param sub_element_name: see doc of base class
        :param value_class: Default value class if no xsi:type is found
        :param cls_getter: function that returns a class for xsi:type QName
        :param ns_helper: name space helper that knows current prefixes
        :param is_optional: see doc of base class
        """
        super().__init__(sub_element_name, ClassCheckConverter(value_class), is_optional=is_optional)
        self.value_class = value_class
        self._cls_getter = cls_getter
        self._ns_helper = ns_helper

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        value = self._default_py_value
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            node_type_str = sub_node.get(QN_TYPE)
            if node_type_str is not None:
                node_type = text_to_qname(node_type_str, node.nsmap)
                value_class = self._cls_getter(node_type)
            else:
                value_class = self.value_class
            value = value_class.from_node(sub_node)
        except ElementNotFoundError:
            pass
        return value

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            py_value = self._default_py_value

        if py_value is None:
            if not self.is_optional:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
                etree.SubElement(node, self._sub_element_name, nsmap=node.nsmap)
        else:
            self.remove_sub_element(node)
            sub_node = py_value.mk_node(self._sub_element_name, self._ns_helper, node)
            if py_value.NODETYPE != self.value_class.NODETYPE:
                # set xsi type
                sub_node.set(QN_TYPE, docname_from_qname(py_value.NODETYPE, node.nsmap))


class _ElementListProperty(_ElementBase, ABC):
    def __get__(self, instance, owner):  # noqa: ANN001
        """Return a python value, uses the locally stored value."""
        if instance is None:  # if called via class
            return self
        try:
            return getattr(instance, self._local_var_name)
        except AttributeError:
            setattr(instance, self._local_var_name, [])
            return getattr(instance, self._local_var_name)

    def __set__(self, instance: Any, py_value: Any):
        if isinstance(py_value, tuple):
            py_value = list(py_value)
        super().__set__(instance, py_value)

    def init_instance_data(self, instance: Any):
        setattr(instance, self._local_var_name, [])

    def update_from_node(self, instance: Any, node: xml_utils.LxmlElement):
        """Update instance data with data from node.

        This method is used internally and should not be called by application.
        :param instance:the instance that has the property as member
        :param node:the etree node that provides the value
        :return:
        """
        value: list | None = self.get_py_value_from_node(instance, node)
        if value is not None:
            setattr(instance, self._local_var_name, value)


class SubElementListProperty(_ElementListProperty):
    """SubElementListProperty is  a list of values that have an "as_etree_node" method.

    Used if maxOccurs="Unbounded" in BICEPS_ParticipantModel.
    """

    def __init__(self, sub_element_name: etree.QName | None, value_class: type[XMLTypeBase], is_optional: bool = True):
        super().__init__(sub_element_name, ListConverter(ClassCheckConverter(value_class)), is_optional=is_optional)
        self.value_class = value_class

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        objects = []
        try:
            nodes = node.findall(self._sub_element_name)
            for _node in nodes:
                value_class = self.value_class.value_class_from_node(_node)
                value = value_class.from_node(_node)
                objects.append(value)
        except ElementNotFoundError:
            pass
        return objects

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = self._default_py_value

        if py_value is not None:
            for val in py_value:
                sub_node = val.as_etree_node(self._sub_element_name, node.nsmap, node)
                if (
                    hasattr(val, 'NODETYPE')
                    and hasattr(self.value_class, 'NODETYPE')
                    and val.NODETYPE != self.value_class.NODETYPE
                ):
                    # set xsi type
                    sub_node.set(QN_TYPE, docname_from_qname(val.NODETYPE, node.nsmap))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} datatype {self.value_class.__name__} in subelement {self._sub_element_name}'


class ContainerListProperty(_ElementListProperty):
    """ContainerListProperty is a property with a list of elements, each supports xsi:type information.

    Used if maxOccurs="Unbounded" in BICEPS_ParticipantModel.
    """

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        value_class: type[ContainerBase],
        cls_getter: Callable[[etree.QName], type],
        ns_helper: NamespaceHelper,
        is_optional: bool = True,
    ):
        """Construct a list of Containers.

        :param sub_element_name: see doc of base class
        :param value_class: Default value class if no xsi:type is found
        :param cls_getter: function that returns a class for xsi:type QName
        :param ns_helper: name space helper that knows current prefixes
        :param is_optional: see doc of base class
        """
        super().__init__(sub_element_name, ListConverter(ClassCheckConverter(value_class)), is_optional=is_optional)
        self.value_class = value_class
        self._cls_getter = cls_getter
        self._ns_helper = ns_helper

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        objects = []
        try:
            nodes = node.findall(self._sub_element_name)
            for _node in nodes:
                node_type_str = _node.get(QN_TYPE)
                if node_type_str is not None:
                    node_type = text_to_qname(node_type_str, _node.nsmap)
                    value_class = self._cls_getter(node_type)
                else:
                    value_class = self.value_class
                value = value_class.from_node(_node)
                objects.append(value)
        except ElementNotFoundError:
            pass
        return objects

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = self._default_py_value

        nodes = node.findall(self._sub_element_name)
        for _node in nodes:
            node.remove(_node)
        # ... and create new ones
        if py_value is not None:
            for val in py_value:
                sub_node = val.mk_node(self._sub_element_name, self._ns_helper, node)
                if val.NODETYPE != self.value_class.NODETYPE:
                    # set xsi type
                    sub_node.set(QN_TYPE, docname_from_qname(val.NODETYPE, node.nsmap))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} datatype {self.value_class.__name__} in subelement {self._sub_element_name}'


class SubElementTextListProperty(_ElementListProperty):
    """SubElementTextListProperty represents a list of strings.

    On xml side every string is a text of a sub element.
    """

    def __init__(self, sub_element_name: etree.QName | None, value_class: Any, is_optional: bool = True):
        super().__init__(sub_element_name, ListConverter(ClassCheckConverter(value_class)), is_optional=is_optional)

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        nodes = node.findall(self._sub_element_name)
        return [_node.text for _node in nodes]

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            py_value = self._default_py_value

        if py_value is None or len(py_value) == 0:
            return

        nodes = node.findall(self._sub_element_name)
        for _node in nodes:
            node.remove(_node)
        # ... and create new ones
        for val in py_value:
            child = etree.SubElement(node, self._sub_element_name)
            try:
                child.text = val
            except TypeError as ex:
                # re-raise with better info about data
                raise TypeError(f'{ex} in {self}') from ex  # noqa: EM102

    def __str__(self) -> str:
        return f'{self.__class__.__name__} in sub-element {self._sub_element_name}'


class SubElementStringListProperty(SubElementTextListProperty):
    """SubElementStringListProperty represents a list of strings.

    On xml side every string is a text of a sub element.
    """

    def __init__(self, sub_element_name: etree.QName | None, is_optional: bool = True):
        super().__init__(sub_element_name, str, is_optional=is_optional)


class SubElementHandleRefListProperty(SubElementStringListProperty):
    """Represents a list of Handles."""


class SubElementWithSubElementListProperty(SubElementProperty):
    """Class represents an optional Element that is only present if its value class is not empty.

    value_class must have an is_empty method.
    """

    def __init__(self, sub_element_name: etree.QName | None, default_py_value: Any, value_class: type[XMLTypeBase]):
        assert hasattr(value_class, 'is_empty')
        super().__init__(sub_element_name, default_py_value=default_py_value, value_class=value_class)

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            py_value = self._default_py_value

        if py_value is None or py_value.is_empty():
            return
        self.remove_sub_element(node)
        py_value.as_etree_node(self._sub_element_name, node.nsmap, node)  # creates a sub-node

    def __set__(self, instance: Any, py_value: Any):
        if isinstance(py_value, self.value_class):
            super().__set__(instance, py_value)
        else:
            raise ApiUsageError(f'do not set {self._sub_element_name} directly, use child member!')  # noqa: EM102


class AnyEtreeNodeListProperty(_ElementListProperty):
    """class represents a list of lxml elements."""

    def __init__(self, sub_element_name: etree.QName | None, is_optional: bool = True):
        super().__init__(
            sub_element_name,
            ListConverter(ClassCheckConverter(xml_utils.LxmlElement)),
            is_optional=is_optional,
        )

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        objects = []
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            if sub_node is None:
                return []
            return sub_node[:]
        except ElementNotFoundError:
            return objects

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:
            py_value = None

        if py_value is None or len(py_value) == 0:
            if self.is_optional:
                self.remove_sub_element(node)
            return

        sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
        sub_node.extend(py_value)

    def __str__(self) -> str:
        return f'{self.__class__.__name__} in sub-element {self._sub_element_name}'


class NodeTextListProperty(_ElementListProperty):
    """The handled data is a list of words (string without whitespace). The xml text is the joined list of words."""

    def __init__(self, sub_element_name: etree.QName | None, value_class: Any, is_optional: bool = False):
        super().__init__(sub_element_name, ListConverter(ClassCheckConverter(value_class)), is_optional=is_optional)

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> list[str] | None:  # noqa: ARG002
        """Read value from node.

        If the expected node does not exist, return _default_py_value (usually None).
        If the node exists, but the text is None, return an empty list.
        """
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            if sub_node.text is not None:
                return sub_node.text.split()
        except ElementNotFoundError:
            return self._default_py_value
        else:
            return []

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None

        if py_value is None:
            if not self._sub_element_name:
                # update text of this element
                node.text = ''
            elif self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            else:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
                sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
                sub_node.text = None
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            sub_node.text = ' '.join(py_value)


class NodeTextQNameListProperty(_ElementListProperty):
    """The handled data is a list of qualified names.

    The xml text is the joined list of qnames in the form prefix:localname.
    """

    def __init__(self, sub_element_name: etree.QName | None, is_optional: bool = False):
        super().__init__(sub_element_name, ListConverter(ClassCheckConverter(etree.QName)), is_optional=is_optional)

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> list[xml_utils.QName] | None:  # noqa: ARG002
        """Read value from node."""
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
        except ElementNotFoundError:
            return self._default_py_value or []
        if sub_node is None:
            return None
        if sub_node.text is not None:
            return [text_to_qname(q_name_string, sub_node.nsmap) for q_name_string in sub_node.text.split()]
        return None

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = None

        if py_value is None:
            if not self._sub_element_name:
                # update text of this element
                node.text = ''
            elif self.is_optional:
                sub_node = node.find(self._sub_element_name)
                if sub_node is not None:
                    node.remove(sub_node)
            else:
                if MANDATORY_VALUE_CHECKING and not self.is_optional:
                    raise ValueError(f'mandatory value {self._sub_element_name} missing')  # noqa: EM102
                sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
                sub_node.text = None
        else:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
            tmp = []
            for q_name in py_value:
                # by setting each qname as text, namespace prefixes are generated automatically
                sub_node.text = q_name
                tmp.append(sub_node.text)
            sub_node.text = ' '.join(tmp)


class DateOfBirthProperty(_ElementBase):
    """DateOfBirthProperty represents the DateOfBirth type of BICEPS.

        <xsd:simpleType>
            <xsd:union memberTypes="xsd:dateTime xsd:date xsd:gYearMonth xsd:gYear"/>
        </xsd:simpleType>
    xsd:dateTime is YYYY-MM-DDThh:mm:ss.sss
    xsd:date is YYYY-MM-DD format. All components are required
    xsd:gYearMonth is YYYY-MM
    xsd:gYear is YYYY
    If the timepoint of birth matters, the value SHALL be populated with a time zone.

    Time zone info can be provided:
       UTC can be specified by appending a Z character, e.g. 2002-09-24Z
       other timezones by adding a positive or negative time behind the date, e.g. 2002.09-24-06:00, 2002-09-24+06:00
    xsd:time is hh:mm:ss format, e.g. 9:30:10, 9:30:10.5. All components are required.
    Time zone handling is identical to date type

    The corresponding Python types are datetime.Date (=> not time point available)
    or datetime.Datetime (with time point attribute).
    """

    def __init__(
        self,
        sub_element_name: etree.QName | None,
        default_py_value: Any = None,
        implied_py_value: Any = None,
        is_optional: bool = True,
    ):
        super().__init__(
            sub_element_name,
            ClassCheckConverter(datetime, date),
            default_py_value,
            implied_py_value,
            is_optional,
        )

    def get_py_value_from_node(self, instance: Any, node: xml_utils.LxmlElement) -> Any:  # noqa: ARG002
        """Read value from node."""
        try:
            sub_node = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=False)
            if sub_node is not None:
                date_string = sub_node.text
                return isoduration.parse_date_time(date_string)
        except ElementNotFoundError:
            pass
        return None

    def update_xml_value(self, instance: Any, node: xml_utils.LxmlElement):
        """Write value to node."""
        try:
            py_value = getattr(instance, self._local_var_name)
        except AttributeError:  # set to None (it is in the responsibility of the called method to do the right thing)
            py_value = self._default_py_value

        if py_value is None:
            self.remove_sub_element(node)
            return

        date_string = py_value if isinstance(py_value, str) else self._mk_datestring(py_value)
        sub_element = self._get_element_by_child_name(node, self._sub_element_name, create_missing_nodes=True)
        sub_element.text = date_string

    @staticmethod
    def mk_value_object(date_string: str) -> isoduration.DateTypeUnion | None:
        """Parse isoduration string."""
        return isoduration.parse_date_time(date_string)

    @staticmethod
    def _mk_datestring(date_object: date | datetime | isoduration.GYear | isoduration.GYearMonth | None) -> str:
        """Create isoduration string."""
        return isoduration.date_time_string(date_object)



from sdc11073.consumer.consumerimpl import SdcConsumer
from sdc11073.consumer.serviceclients.containmenttreeservice import CTreeServiceClient
from sdc11073.consumer.serviceclients.contextservice import ContextServiceClient
from sdc11073.consumer.serviceclients.descriptioneventservice import DescriptionEventClient
from sdc11073.consumer.serviceclients.getservice import GetServiceClient
from sdc11073.consumer.serviceclients.localizationservice import LocalizationServiceClient
from sdc11073.consumer.serviceclients.setservice import SetServiceClient
from sdc11073.consumer.serviceclients.stateeventservice import StateEventClient
from sdc11073.consumer.serviceclients.waveformservice import WaveformClient

__all__ = [
    'SdcConsumer',
    'CTreeServiceClient',
    'ContextServiceClient',
    'DescriptionEventClient',
    'GetServiceClient',
    'LocalizationServiceClient',
    'SetServiceClient',
    'StateEventClient',
    'WaveformClient',
]



"""Implementation the types described in ws-addressing core from 2006.

More information can be found at
- https://www.w3.org/TR/2006/REC-ws-addr-core-20060509/
- https://www.w3.org/TR/2006/REC-ws-addr-soap-20060509/.
"""

from __future__ import annotations

import copy
import uuid
from typing import TYPE_CHECKING

from sdc11073.namespaces import default_ns_helper as nsh
from sdc11073.xml_types import xml_structure as struct
from sdc11073.xml_types.basetypes import ElementWithText, XMLTypeBase

if TYPE_CHECKING:
    from lxml import etree

    from sdc11073 import xml_utils

_is_reference_parameter = nsh.WSA.tag('IsReferenceParameter')


class EndpointReferenceType(XMLTypeBase):
    """The wsa:EndpointReferenceType type is used wherever a Web service endpoint is referenced.

    The following describes the contents of this type:
    <wsa:EndpointReference>
        <wsa:Address>xs:anyURI</wsa:Address>
        <wsa:ReferenceParameters>xs:any*</wsa:ReferenceParameters> ?
        <wsa:Metadata>xs:any*</wsa:Metadata>?
    </wsa:EndpointReference>
    """

    Address = struct.AnyUriTextElement(nsh.WSA.tag('Address'))
    ReferenceParameters: list[xml_utils.LxmlElement] | None = (
        struct.AnyEtreeNodeListProperty(nsh.WSA.tag('ReferenceParameters'), is_optional=True))
    Metadata: list[xml_utils.LxmlElement] | None = (
        struct.AnyEtreeNodeListProperty(nsh.WSA.tag('Metadata'), is_optional=True))
    _props = ('Address', 'ReferenceParameters', 'Metadata')


class RelatesTo(ElementWithText):
    """Contributes one abstract [relationship] property value."""

    RelationshipType: str | None = struct.AnyUriTextElement(
        nsh.WSA.tag('RelationshipType'),
        is_optional=True,
        implied_py_value='http://www.w3.org/2005/08/addressing/reply')
    _props = ('RelationshipType',)


class HeaderInformationBlock(XMLTypeBase):
    """HeaderInformationBlock contains data that ws-addressing requires in soap header.

    <wsa:To>xs:anyURI</wsa:To> ?
    <wsa:From>wsa:EndpointReferenceType</wsa:From> ?
    <wsa:ReplyTo>wsa:EndpointReferenceType</wsa:ReplyTo> ?
    <wsa:FaultTo>wsa:EndpointReferenceType</wsa:FaultTo> ?
    <wsa:Action>xs:anyURI</wsa:Action>
    <wsa:MessageID>xs:anyURI</wsa:MessageID> ?
    <wsa:RelatesTo RelationshipType="xs:anyURI"?>xs:anyURI</wsa:RelatesTo> *

    in soap this is xs:any with wsa:IsReferenceParameter='true' as list
    <wsa:ReferenceParameters>xs:any*</wsa:ReferenceParameters> ?
    """

    To: str | None = struct.AnyUriTextElement(nsh.WSA.tag('To'), is_optional=True)
    From: EndpointReferenceType | None = struct.SubElementProperty(nsh.WSA.tag('From'),
                                                                   value_class=EndpointReferenceType,
                                                                   is_optional=True)
    ReplyTo: EndpointReferenceType | None = struct.SubElementProperty(nsh.WSA.tag('ReplyTo'),
                                                                      value_class=EndpointReferenceType,
                                                                      is_optional=True)
    FaultTo: EndpointReferenceType | None = struct.SubElementProperty(nsh.WSA.tag('FaultTo'),
                                                                      value_class=EndpointReferenceType,
                                                                      is_optional=True)
    Action: str = struct.AnyUriTextElement(nsh.WSA.tag('Action'))
    # note: ws-addressing declares MessageId as optional, but it is required for ws-discovery
    MessageID: str | None = struct.AnyUriTextElement(nsh.WSA.tag('MessageID'), is_optional=True)
    # note: following the standard would require RelatesTo to be a list, but ws-discovery requires it to be 0..1
    RelatesTo = struct.SubElementProperty(nsh.WSA.tag('RelatesTo'),
                                          value_class=RelatesTo,
                                          is_optional=True)

    _props = ('To', 'From', 'ReplyTo', 'FaultTo', 'Action', 'MessageID', 'RelatesTo')

    def __init__(self,  # noqa: PLR0913
                 action: str | None = None,
                 message_id: str | None = None,
                 addr_to: str | None = None,
                 relates_to: str | None = None,
                 addr_from: str | None = None,
                 reference_parameters: list[xml_utils.LxmlElement] | None = None,
                 relationship_type: str | None = None):
        super().__init__()
        self.Action = action
        self.MessageID = message_id or uuid.uuid4().urn
        if addr_to is not None:
            self.To = addr_to
        if relates_to is not None:
            self.RelatesTo = RelatesTo()
            self.RelatesTo.text = relates_to
            if relationship_type is not None:
                self.RelatesTo.RelationshipType = relationship_type
        if addr_from is not None:
            self.From = EndpointReferenceType()
            self.From.Address = addr_from
        self.reference_parameters: list[xml_utils.LxmlElement] = reference_parameters or []

    def mk_reply_header_block(self,
                              action: str | None = None,
                              message_id: str | None = None,
                              addr_to: str | None = None) -> HeaderInformationBlock:
        """Create a HeaderInformationBlock with RelatesTo information of self."""
        reply_address = HeaderInformationBlock(action, message_id, addr_to)
        reply_address.RelatesTo = RelatesTo()
        reply_address.RelatesTo.text = self.MessageID
        reply_address.Action = action
        return reply_address

    def as_etree_node(self,
                      q_name: etree.QName, ns_map: dict[str, str],
                      parent_node: xml_utils.LxmlElement | None = None) -> xml_utils.LxmlElement:
        """Create etree Element form instance data."""
        node = super().as_etree_node(q_name, ns_map, parent_node)
        for param in self.reference_parameters:
            tmp = copy.deepcopy(param)
            tmp.set(_is_reference_parameter, 'true')
            node.append(tmp)
        return node

    @classmethod
    def from_node(cls, node: xml_utils.LxmlElement) -> HeaderInformationBlock:
        """Create HeaderInformationBlock from etree element."""
        obj = cls()
        obj.update_from_node(node)
        # collect reference parameter child nodes
        for child in node:
            is_reference_parameter = child.attrib.get(_is_reference_parameter, 'false')
            if is_reference_parameter.lower() == 'true':
                obj.reference_parameters.append(child)
        return obj



"""Implementation of entities for EntityProviderMdib."""
from __future__ import annotations

import copy
import uuid
from typing import TYPE_CHECKING, TypeVar, Union

from lxml.etree import QName

from sdc11073.mdib.containerbase import ContainerBase
from sdc11073.mdib.entityprotocol import EntityProtocol, MultiStateEntityProtocol
from sdc11073.namespaces import QN_TYPE, text_to_qname
from sdc11073.xml_types import pm_qnames
from sdc11073.xml_types.pm_types import CodedValue

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorContainer
    from sdc11073.mdib.statecontainers import AbstractMultiStateContainer, AbstractStateContainer
    from sdc11073.xml_utils import LxmlElement

    from .entity_consumermdib import EntityConsumerMdib
    from .entity_providermdib import EntityProviderMdib

# Many types are fixed in schema. This table maps from tag in Element to its type
_static_type_lookup = {
    pm_qnames.Mds: pm_qnames.MdsDescriptor,
    pm_qnames.Vmd: pm_qnames.VmdDescriptor,
    pm_qnames.Channel: pm_qnames.ChannelDescriptor,
    pm_qnames.AlertSystem: pm_qnames.AlertSystemDescriptor,
    pm_qnames.AlertCondition: pm_qnames.AlertConditionDescriptor,
    pm_qnames.AlertSignal: pm_qnames.AlertSignalDescriptor,
    pm_qnames.Sco: pm_qnames.ScoDescriptor,
    pm_qnames.SystemContext: pm_qnames.SystemContextDescriptor,
    pm_qnames.PatientContext: pm_qnames.PatientContextDescriptor,
    pm_qnames.LocationContext: pm_qnames.LocationContextDescriptor,
    pm_qnames.EnsembleContext: pm_qnames.EnsembleContextDescriptor,
    pm_qnames.Clock: pm_qnames.ClockDescriptor,
    pm_qnames.Battery: pm_qnames.BatteryDescriptor,
}


def get_xsi_type(element: LxmlElement) -> QName:
    """Return the BICEPS type of an element.

    If there is a xsi:type entry, this specifies the type.
    If not, the tag is used to determine the type.
    """
    xsi_type_str = element.attrib.get(QN_TYPE)

    if xsi_type_str:
        return text_to_qname(xsi_type_str, element.nsmap)
    _xsi_type = QName(element.tag)
    try:
        return _static_type_lookup[_xsi_type]
    except KeyError as err:  # pragma: no cover
        raise KeyError(str(_xsi_type)) from err


class _XmlEntityBase:
    """A descriptor element and some info about it for easier access."""

    def __init__(self,
                 parent_handle: str | None,
                 source_mds: str | None,
                 node_type: QName,
                 descriptor: LxmlElement):
        self.parent_handle = parent_handle
        self.source_mds = source_mds
        self.node_type = node_type  # name of descriptor type
        self._descriptor = descriptor
        self.coded_value: CodedValue | None = None
        self.descriptor = descriptor  # setter updates self._descriptor and self.coded_value

    @property
    def descriptor(self) -> LxmlElement:
        return self._descriptor

    @descriptor.setter
    def descriptor(self, new_descriptor: LxmlElement):
        self._descriptor = new_descriptor
        type_node = self.descriptor.find(pm_qnames.Type)
        if type_node is not None:
            self.coded_value = CodedValue.from_node(type_node)
        else:
            self.coded_value = None

    def __str__(self):
        return f'{self.__class__.__name__} {self.node_type.localname} handle={self._descriptor.get("Handle")}'


class XmlEntity(_XmlEntityBase):
    """Groups descriptor and state."""

    def __init__(self,
                 parent_handle: str | None,
                 source_mds: str | None,
                 node_type: QName,
                 descriptor: LxmlElement,
                 state: LxmlElement | None):
        super().__init__(parent_handle, source_mds, node_type, descriptor)
        self.state = state

    @property
    def is_multi_state(self) -> bool:
        """Return False because this is not a multi state entity."""
        return False

    def mk_entity(self, mdib: EntityConsumerMdib) -> ConsumerEntity:
        """Return a corresponding entity with containers."""
        return ConsumerEntity(self, mdib)


class XmlMultiStateEntity(_XmlEntityBase):
    """Groups descriptor and list of multi-states."""

    def __init__(self,
                 parent_handle: str | None,
                 source_mds: str | None,
                 node_type: QName,
                 descriptor: LxmlElement,
                 states: list[LxmlElement]):
        super().__init__(parent_handle, source_mds, node_type, descriptor)
        self.states: dict[str, LxmlElement] = {node.get('Handle'): node for node in states}

    @property
    def is_multi_state(self) -> bool:
        """Return True because this is a multi state entity."""
        return True

    def mk_entity(self, mdib: EntityConsumerMdib) -> ConsumerMultiStateEntity:
        """Return a corresponding entity with containers."""
        return ConsumerMultiStateEntity(self, mdib)


class ConsumerEntityBase:
    """A descriptor container and a weak reference to the corresponding xml entity."""

    def __init__(self,
                 source: XmlEntity | XmlMultiStateEntity,
                 mdib: EntityConsumerMdib,  # needed if a new state needs to be added
                 ):
        self._mdib: EntityConsumerMdib = mdib

        cls = mdib.sdc_definitions.data_model.get_descriptor_container_class(source.node_type)
        if cls is None:  # pragma: no cover
            msg = f'do not know how to make container from {source.node_type!s}'
            raise ValueError(msg)
        handle = source.descriptor.get('Handle')
        self.descriptor: AbstractDescriptorContainer = cls(handle, parent_handle=source.parent_handle)
        self.descriptor.update_from_node(source.descriptor)
        if source.source_mds is not None:
            self.descriptor.set_source_mds(source.source_mds)
        self.source_mds = source.source_mds

    @property
    def handle(self) -> str:
        """Return the handle of the descriptor."""
        return self.descriptor.Handle

    @property
    def parent_handle(self) -> str | None:
        """Return the parent handle of the descriptor."""
        return self.descriptor.parent_handle

    @property
    def node_type(self) -> QName:
        """Return the node type of the descriptor."""
        return self.descriptor.NODETYPE

    def __str__(self):
        return f'{self.__class__.__name__} {self.node_type} handle={self.handle}'


class ConsumerEntity(ConsumerEntityBase, EntityProtocol):
    """Groups descriptor container and state container."""

    def __init__(self,
                 source: XmlEntity,
                 mdib: EntityConsumerMdib):
        super().__init__(source, mdib)
        self.state: AbstractStateContainer | None = None
        if source.state is not None:
            self._mk_state(source.state)

    def update(self):
        """Update the entity from current data in mdib."""
        xml_entity = self._mdib.internal_entities.get(self.handle)
        if xml_entity is None:
            raise ValueError('entity no longer exists in mdib')
        if int(xml_entity.descriptor.get('DescriptorVersion', '0')) != self.descriptor.DescriptorVersion:
            self.descriptor.update_from_node(xml_entity.descriptor)
        if xml_entity.state is not None:
            if self.state is None:
                self._mk_state(xml_entity.state)
            elif int(xml_entity.state.get('StateVersion', '0')) != self.state.StateVersion:
                self.state.update_from_node(xml_entity.state)

    def _mk_state(self, lxml_state: LxmlElement):
        cls = self._mdib.sdc_definitions.data_model.get_state_container_class(self.descriptor.STATE_QNAME)
        self.state = cls(self.descriptor)
        self.state.update_from_node(lxml_state)


class ConsumerMultiStateEntity(ConsumerEntityBase, MultiStateEntityProtocol):
    """Groups descriptor container and list of multi-state containers."""

    def __init__(self,
                 source: XmlMultiStateEntity,
                 mdib: EntityConsumerMdib):
        super().__init__(source, mdib)
        self.states: dict[str, AbstractMultiStateContainer] = {}
        for handle, state in source.states.items():
            state_type = get_xsi_type(state)
            cls = mdib.sdc_definitions.data_model.get_state_container_class(state_type)
            state_container = cls(self.descriptor)
            state_container.update_from_node(state)
            self.states[handle] = state_container

    def update(self):
        """Update the entity from current data in mdib."""
        xml_entity = self._mdib.internal_entities.get(self.handle)
        if xml_entity is None:
            raise ValueError('entity no longer exists in mdib')
        if int(xml_entity.descriptor.get('DescriptorVersion', '0')) != self.descriptor.DescriptorVersion:
            self.descriptor.update_from_node(xml_entity.descriptor)

        for handle, xml_state in xml_entity.states.items():
            existing_state = self.states.get(handle)
            if existing_state is None:
                # create new state
                xsi_type = get_xsi_type(xml_state)
                cls = self._mdib.sdc_definitions.data_model.get_state_container_class(xsi_type)
                state_container = cls(self.descriptor)
                state_container.update_from_node(xml_state)
                self.states[handle] = state_container
            elif existing_state.StateVersion != int(xml_state.get('StateVersion', '0')):
                existing_state.update_from_node(xml_state)

        # delete states that are no longer in xml_entity
        for handle in list(self.states.keys()):
            if handle not in xml_entity.states:
                self.states.pop(handle)

    def new_state(self, state_handle: str | None = None) -> AbstractMultiStateContainer:
        """Create a new state.

        The new state has handle of descriptor container as handle.
        If this new state is used as a proposed context state in SetContextState operation, this means a new
        state shall be created on providers side.
        """
        if state_handle in self.states:  # pragma: no cover
            msg = f'State handle {state_handle} already exists in {self.__class__.__name__}, handle = {self.handle}'
            raise ValueError(msg)
        cls = self._mdib.data_model.get_state_container_class(self.descriptor.STATE_QNAME)
        state = cls(descriptor_container=self.descriptor)
        state.Handle = state_handle or self.handle
        return state


ConsumerEntityType = Union[ConsumerEntity, ConsumerMultiStateEntity]
ConsumerInternalEntityType = Union[XmlEntity, XmlMultiStateEntity]


##############  provider ##########################

class ProviderInternalEntityBase:
    """A descriptor element and some info about it for easier access."""

    def __init__(self, descriptor: AbstractDescriptorContainer):
        self.descriptor = descriptor

    @property
    def handle(self) -> str:
        """Return the handle of the descriptor."""
        return self.descriptor.Handle

    @property
    def parent_handle(self) -> str | None:
        """Return the parent handle of the descriptor."""
        return self.descriptor.parent_handle

    @property
    def source_mds(self) -> str:
        """Return the source mds of the descriptor."""
        return self.descriptor.source_mds

    @property
    def node_type(self) -> QName:
        """Return the node type of the descriptor."""
        return self.descriptor.NODETYPE

    def __str__(self):
        return f'{self.__class__.__name__} {self.node_type.localname} handle={self.handle}'


class ProviderInternalEntity(ProviderInternalEntityBase):
    """Groups descriptor and state."""

    def __init__(self,
                 descriptor: AbstractDescriptorContainer,
                 state: AbstractStateContainer | None):
        super().__init__(descriptor)
        self._state = state
        if state is not None:
            self._state.descriptor_container = self.descriptor

    @property
    def state(self) -> AbstractStateContainer | None:
        """Return the state member of the entity."""
        return self._state

    @state.setter
    def state(self, new_state: AbstractStateContainer):
        self._state = new_state
        self._state.descriptor_container = self.descriptor

    @property
    def is_multi_state(self) -> bool:
        """Return False because this is not a multi state entity."""
        return False

    def mk_entity(self, mdib: EntityProviderMdib) -> ProviderEntity:
        """Return a corresponding entity with containers."""
        return ProviderEntity(self, mdib)


class ProviderInternalMultiStateEntity(ProviderInternalEntityBase):
    """Groups descriptor and list of multi-states."""

    def __init__(self,
                 descriptor: AbstractDescriptorContainer,
                 states: list[AbstractMultiStateContainer]):
        super().__init__(descriptor)
        self.states = {state.Handle: state for state in states}

    @property
    def is_multi_state(self) -> bool:
        """Return True because this is a multi state entity."""
        return True

    def mk_entity(self, mdib: EntityProviderMdib) -> ProviderMultiStateEntity:
        """Return a corresponding entity with containers."""
        return ProviderMultiStateEntity(self, mdib)


T = TypeVar('T', bound=ContainerBase)
def _mk_copy(original: T) -> T:
    """Return a deep copy of original without node member."""
    node, original.node = original.node, None
    copied = copy.deepcopy(original)
    original.node = node
    return copied


class ProviderEntityBase:
    """A descriptor container and a weak reference to the corresponding xml entity."""

    def __init__(self,
                 source: ProviderInternalEntity | ProviderInternalMultiStateEntity,
                 mdib: EntityProviderMdib):
        self._mdib = mdib
        self.descriptor = _mk_copy(source.descriptor)
        self.source_mds = source.source_mds

    @property
    def handle(self) -> str:
        """Return the handle of the descriptor."""
        return self.descriptor.Handle

    @property
    def parent_handle(self) -> str | None:
        """Return the parent handle of the descriptor."""
        return self.descriptor.parent_handle

    @property
    def node_type(self) -> QName:
        """Return the node type of the descriptor."""
        return self.descriptor.NODETYPE

    def __str__(self):
        return f'{self.__class__.__name__} {self.node_type.localname} handle={self.handle}'


class ProviderEntity(ProviderEntityBase):
    """Groups descriptor container and state container."""

    def __init__(self,
                 source: ProviderInternalEntity,
                 mdib: EntityProviderMdib):
        super().__init__(source, mdib)
        self.state: AbstractStateContainer | None = None
        if source.state is not None:
            self.state = _mk_copy(source.state)

    @property
    def is_multi_state(self) -> bool:
        """Return False because this is not a multi state entity."""
        return False

    def update(self):
        """Update from internal entity."""
        source_entity = self._mdib.internal_entities.get(self.handle)
        if source_entity is None: # pragma: no cover
            msg = f'entity {self.handle} no longer exists in mdib'
            raise ValueError(msg)
        self.descriptor.update_from_other_container(source_entity.descriptor)
        self.state = _mk_copy(source_entity.state)


class ProviderMultiStateEntity(ProviderEntityBase):
    """Groups descriptor container and list of multi-state containers."""

    def __init__(self,
                 source: ProviderInternalMultiStateEntity,
                 mdib: EntityProviderMdib):
        super().__init__(source, mdib)
        self.states: dict[str, AbstractMultiStateContainer] = {st.Handle: _mk_copy(st) for st in source.states.values()}

    @property
    def is_multi_state(self) -> bool:
        """Return True because this is a multi state entity."""
        return True

    def update(self):
        """Update from internal entity."""
        source_entity = self._mdib.internal_entities.get(self.handle)
        if source_entity is None:  # pragma: no cover
            msg = f'entity {self.handle} no longer exists in mdib'
            raise ValueError(msg)
        self.descriptor.update_from_other_container(source_entity.descriptor)
        for handle, src_state in source_entity.states.items():
            dest_state = self.states.get(handle)
            if dest_state is None:
                self.states[handle] = _mk_copy(src_state)
            else:
                dest_state.update_from_other_container(src_state)
        # remove states that are no longer present is source_entity
        for handle in list(self.states.keys()):
            if handle not in source_entity.states:
                self.states.pop(handle)

    def new_state(self, state_handle: str | None = None) -> AbstractMultiStateContainer:
        """Create a new state."""
        if state_handle in self.states: # pragma: no cover
            msg = f'State handle {state_handle} already exists in {self.__class__.__name__}, handle = {self.handle}'
            raise ValueError(msg)
        cls = self._mdib.data_model.get_state_container_class(self.descriptor.STATE_QNAME)
        state = cls(descriptor_container=self.descriptor)
        state.Handle = state_handle or uuid.uuid4().hex
        self.states[state.Handle] = state
        return state


ProviderInternalEntityType = Union[ProviderInternalEntity, ProviderInternalMultiStateEntity]
ProviderEntityType = Union[ProviderEntity, ProviderMultiStateEntity]
AnyProviderEntityType = Union[ProviderInternalEntity, ProviderInternalMultiStateEntity ]



"""The module declares several protocols that are implemented by transactions.

Only these protocols shall be used, the old way of transactions in mdib.transactions should no longer be used.
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Protocol, Union

from .statecontainers import AbstractMultiStateProtocol, AbstractStateProtocol

if TYPE_CHECKING:

    from .descriptorcontainers import AbstractDescriptorProtocol
    from .entityprotocol import EntityProtocol, EntityTypeProtocol, MultiStateEntityProtocol

class TransactionType(Enum):
    """The different kinds of transactions.

    Each type results in a different report.
    """

    descriptor = 1
    metric = 2
    alert = 3
    component = 4
    context = 5
    operational = 6
    rt_sample = 7


class TransactionResultProtocol(Protocol): # pragma: no cover
    """TransactionResult contains all state and descriptors that were modified in the transaction.

    The states and descriptors are used to create the notification(s) that keep the consumers up to date.
    """

    descr_updated: list[AbstractDescriptorProtocol]
    descr_created: list[AbstractDescriptorProtocol]
    descr_deleted: list[AbstractDescriptorProtocol]
    metric_updates: list[AbstractStateProtocol]
    alert_updates: list[AbstractStateProtocol]
    comp_updates = list[AbstractStateProtocol]
    ctxt_updates = list[AbstractMultiStateProtocol]
    op_updates = list[AbstractStateProtocol]
    rt_updates = list[AbstractStateProtocol]

    has_descriptor_updates: bool
    new_mdib_version: int

    def all_states(self) -> list[AbstractStateProtocol]:
        """Return all states in this transaction."""

class TransactionItemProtocol(Protocol): # pragma: no cover
    """A container for the old and the new version of a state or descriptor.

    If old is None, this is an object that is added to mdib.
    If new is None, this is an object that is deleted from mdib.
    If neither old nor new is None, this is an update to an existing object.
    """

    old: AbstractStateProtocol | AbstractDescriptorProtocol | None
    new: AbstractStateProtocol | AbstractDescriptorProtocol | None


@dataclass(frozen=True)
class TransactionItem:
    """Transaction Item with old and new container."""

    old: AbstractStateProtocol | AbstractDescriptorProtocol | None
    new: AbstractStateProtocol | AbstractDescriptorProtocol | None




class AbstractTransactionManagerProtocol(Protocol): # pragma: no cover
    """Interface of a TransactionManager."""

    new_mdib_version: int

    def process_transaction(self, set_determination_time: bool) -> TransactionResultProtocol:
        """Process the transaction."""

    # member variables that are available during a running transaction
    descriptor_updates: dict[str, TransactionItem]
    metric_state_updates: dict[str, TransactionItem]
    alert_state_updates: dict[str, TransactionItem]
    component_state_updates: dict[str, TransactionItem]
    context_state_updates: dict[str, TransactionItem]
    operational_state_updates: dict[str, TransactionItem]
    rt_sample_state_updates: dict[str, TransactionItem]
    error: bool


class EntityDescriptorTransactionManagerProtocol(AbstractTransactionManagerProtocol): # pragma: no cover
    """Entity based transaction manager for modification of descriptors (and associated states).

    The entity based transaction manager protocol can only be used with EntityGetter methods!
    The only working approach is:
        1. Read an entity from mdib with one of the EntityGetter Methods. These methods return a
           copy of the mdib data.
        2. Manipulate the copied data as required
        3. Create a transaction and write entity data back to mdib with write_entity method
    """

    def get_state_transaction_item(self, handle: str) -> TransactionItem | None:
        """If transaction has a state with given handle, return the transaction-item, otherwise None."""

    def transaction_entity(self, descriptor_handle: str) -> EntityTypeProtocol | None:
        """Return the entity in open transaction if it exists.

        The descriptor can already be part of the transaction, and e.g. in pre_commit handlers of role providers
        it can be necessary to have access to it.
        """

    def write_entity(self,
                     entity: EntityTypeProtocol,
                     adjust_version_counter: bool = True):
        """Insert or update an entity (state and descriptor)."""

    def write_entities(self,
                       entities: list[EntityTypeProtocol],
                       adjust_version_counter: bool = True):
        """Insert or update list of entities."""

    def remove_entity(self, entity: EntityTypeProtocol):
        """Remove existing descriptor from mdib."""


class EntityStateTransactionManagerProtocol(AbstractTransactionManagerProtocol): # pragma: no cover
    """Entity based transaction manager for modification of states.

    The entity based transaction manager protocol can only be used with EntityGetter methods!
    The only working approach is:
        1. Read an entity from mdib with one of the EntityGetter Methods. These methods return a
           copy of the mdib data.
        2. Manipulate the copied state as required
        3. Create a transaction and write entity data back to mdib with write_entity method
    """

    def has_state(self, descriptor_handle: str) -> bool:
        """Check if transaction has a state with given handle."""

    def write_entity(self,
                     entity: EntityProtocol,
                     adjust_version_counter: bool = True):
        """Update the state of the entity."""

    def write_entities(self,
                       entities: list[EntityProtocol],
                       adjust_version_counter: bool = True):
        """Update the states of entities."""


class EntityContextStateTransactionManagerProtocol(AbstractTransactionManagerProtocol): # pragma: no cover
    """Entity based transaction manager for modification of context states.

    The entity based transaction manager protocol can only be used with EntityGetter methods!
    The only working approach is:
        1. Read an entity from mdib with one of the EntityGetter Methods. These methods return a
           copy of the mdib data.
        2. Manipulate the copied states as required
        3. Create a descriptor transaction context and write entity data back to mdib with write_entity method
    """

    def write_entity(self, entity: MultiStateEntityProtocol,
                  modified_handles: list[str],
                  adjust_version_counter: bool = True):
        """Insert or update a context state in mdib."""


class DescriptorTransactionManagerProtocol(EntityDescriptorTransactionManagerProtocol): # pragma: no cover
    """The classic Interface of a TransactionManager that modifies descriptors.

    The classic transaction manager protocol can not be used with EntityGetter methods!
    The only working approach is:
        case A: update an existing descriptor:
        1. Start a descriptor transaction context
        2. call get_descriptor. This returns a copy of the descriptor in mdib
           Manipulate the copied descriptor as required
        3. optional: call get_state / get_context_state. This returns a copy of the state in mdib
           Manipulate the copied state as required

        case B: create a descriptor ( not context descriptor):
        1. Start a descriptor transaction context
        2. Create a new descriptor (and state instance if this is not a context state)
        3. Call add_descriptor and add_state

        case C: create a context descriptor:
        1. Start a descriptor transaction context
        2. Create a new descriptor
        3. Call mk_context_state  0... n times to add context states

        In all cases: when the transaction context is left, all before retrieved data is written back to mdib.
    """

    def actual_descriptor(self, descriptor_handle: str) -> AbstractDescriptorProtocol:
        """Look for new or updated descriptor in current transaction and in mdib."""

    def add_descriptor(self,
                       descriptor_container: AbstractDescriptorProtocol,
                       adjust_descriptor_version: bool = True,
                       state_container: AbstractStateProtocol | None = None):
        """Add a new descriptor to mdib."""

    def remove_descriptor(self, descriptor_handle: str):
        """Remove existing descriptor from mdib."""

    def get_descriptor(self, descriptor_handle: str) -> AbstractDescriptorProtocol:
        """Get a descriptor from mdib."""

    def has_state(self, descriptor_handle: str) -> bool:
        """Check if transaction has a state with given handle."""

    def add_state(self, state_container: AbstractStateProtocol, adjust_state_version: bool = True):
        """Add a new state to mdib."""

    def unget_state(self, state_container: AbstractStateProtocol):
        """Forget a state that was provided before by a get_state or add_state call."""

    def get_state(self, descriptor_handle: str) -> AbstractStateProtocol:
        """Read a state from mdib and add it to the transaction."""

    def get_context_state(self, context_state_handle: str) -> AbstractMultiStateProtocol:
        """Read a ContextState from mdib with given state handle."""

    def mk_context_state(self, descriptor_handle: str,
                         context_state_handle: str | None = None,
                         adjust_state_version: bool = True,
                         set_associated: bool = False) -> AbstractMultiStateProtocol:
        """Create a new ContextStateContainer."""


class StateTransactionManagerProtocol(EntityStateTransactionManagerProtocol): # pragma: no cover
    """The classic Interface of a TransactionManager that modifies states (except context states).

    The classic transaction manager protocol can not be used with EntityGetter methods!
    The only working approach is:
        1. Start a descriptor transaction context
        2. call get_state. This returns a copy of the state in mdib
           Manipulate the copied state as required

        When the transaction context is left, all before retrieved data is written back to mdib.
    """

    def has_state(self, descriptor_handle: str) -> bool:
        """Check if transaction has a state with given handle."""

    def get_state_transaction_item(self, handle: str) -> TransactionItemProtocol | None:
        """If transaction has a state with given handle, return the transaction-item, otherwise None."""

    def unget_state(self, state_container: AbstractStateProtocol):
        """Forget a state that was provided before by a get_state or add_state call."""

    def get_state(self, descriptor_handle: str) -> AbstractStateProtocol:
        """Read a state from mdib and add it to the transaction."""


class ContextStateTransactionManagerProtocol(EntityContextStateTransactionManagerProtocol): # pragma: no cover
    """The classic Interface of a TransactionManager that modifies context states.

    The classic transaction manager protocol can not be used with EntityGetter methods!
    The only working approach is:
        1. Start a descriptor transaction context
        2a.Call get_context_state if you want to manipulate an existing context state.
           This returns a copy of the state in mdib. Manipulate the copied state as required.
        2b.Call mk_context_state if you want to create a new context state.
           Manipulate the state as required.

        When the transaction context is left, all before retrieved data is written back to mdib.
    """

    def get_context_state(self, context_state_handle: str) -> AbstractMultiStateProtocol:
        """Read a ContextState from mdib with given state handle."""

    def mk_context_state(self, descriptor_handle: str,
                         context_state_handle: str | None = None,
                         adjust_state_version: bool = True,
                         set_associated: bool = False) -> AbstractMultiStateProtocol:
        """Create a new ContextStateContainer."""

    def disassociate_all(self,
                         context_descriptor_handle: str,
                         ignored_handle: str | None = None) -> list[str]:
        """Disassociate all associated states in mdib for context_descriptor_handle."""

AnyEntityTransactionManagerProtocol = Union[EntityContextStateTransactionManagerProtocol,
                                            EntityStateTransactionManagerProtocol,
                                            EntityDescriptorTransactionManagerProtocol]


AnyTransactionManagerProtocol = Union[ContextStateTransactionManagerProtocol,
                                      StateTransactionManagerProtocol,
                                      DescriptorTransactionManagerProtocol]




""" This module defines an ObservableProperty class.
An ObservablePropery must be declared as class attribute, similar to standard python properties.
You can bind callables to an ObservableProperty. The callable is called when the property value is set.

Example:
-------
>>> class MyBaseClass:
>>>     prop1 = ObservableProperty(21)
>>>     prop2 = ObservableProperty(22)
>>>
>>>     def __init__(self):
>>>         do_something()
>>>
>>> class Observer:
>>>     def onProp1Changed(self, value):
>>>         print 'prop1=', value
>>>     def onProp21Changed(self, value):
>>>         print 'prop2=', value
>>>
>>> actor = MyBaseClass()
>>> observer = Observer()
>>> bind(actor, prop1=observer.onProp1Changed, prop2=observer.onProp2Changed)

>>> actor.prop1=42
< prop1= 42
>>> actor.prop2='Hello World'
< prop2= Hello World

"""
import contextlib
import inspect
import weakref
from contextlib import contextmanager


class WeakRef:
    """ This Weak Ref implementation allows to hold references to bound methods.
    => see http://stackoverflow.com/questions/599430/why-doesnt-the-weakref-work-on-this-bound-method"""

    def __init__(self, item):
        self.reference = None
        self.method = None
        self.instance = None
        try:
            self.method = item.__func__.__name__
            self.instance = weakref.ref(item.__self__)
        except AttributeError:
            self.reference = weakref.ref(item)

    def get_ref(self):
        if self.reference is not None:
            return self.reference()
        instance = self.instance()
        if instance is None:
            return None
        return getattr(instance, self.method)

    def __eq__(self, other):
        try:
            if self.reference is not None:
                return self.reference == other.reference
            return self.method == other.method and self.instance == other.instance
        except AttributeError:
            # other is of an unknown class
            return False


class _ObservableValue:
    """ Implements the basic mechanism for an observable value. """

    def __init__(self, value, fire_only_on_changed_value=True):
        self.value = value
        self._fire_only_on_changed_value = fire_only_on_changed_value
        self._observers = []

    def set_value(self, value):
        if value == self.value and self._fire_only_on_changed_value:
            return
        self.value = value
        obsolete_refs = []
        # now call all listeners. Keep track of obsolete weak references
        for ref in self._observers[:]:  # make a copy of list, content might change during iteration
            try:
                func = ref.get_ref()
            except AttributeError:  # no Weakref instance => strong reference, use ref directly
                func = ref
            if func is None:
                obsolete_refs.append(ref)
            else:
                func(self.value)  # call func
        for ref in obsolete_refs:
            with contextlib.suppress(ValueError):  # e.g. has been deleted by someone else in different thread
                self._observers.remove(ref)

    def bind(self, func):
        self._observers.append(WeakRef(func))

    def strongbind(self, func):
        self._observers.append(func)

    def unbind(self, func):
        func_ref = WeakRef(func)
        for ref in self._observers:
            if ref in (func, func_ref):
                self._observers.remove(ref)
                break

    def unbind_all(self):
        del self._observers[:]


class ObservableProperty:
    """ stores data in parent obj """

    def __init__(self, default_value=None, fire_only_on_changed_value=True):
        self._default_value = default_value
        self._fire_only_on_changed_value = fire_only_on_changed_value

    def _get_instance_data(self, obj):
        # see if we already have a _property_instance_data dictionary injected in obj
        # otherwise inject it
        # pylint: disable=protected-access
        try:
            lookup = obj._property_instance_data
        except AttributeError:
            obj._property_instance_data = {}
            lookup = obj._property_instance_data
        # pylint: enable=protected-access

        # see if we already have a data instance for my property instance and class instance
        # otherwise create one
        try:
            return lookup[self]
        except KeyError:
            lookup[self] = _ObservableValue(self._default_value, self._fire_only_on_changed_value)
            return lookup[self]

    def __get__(self, obj, objtype):
        return self if obj is None else self._get_instance_data(obj).value

    def __set__(self, obj, value):
        if obj is None:
            self._default_value = value
        else:
            self._get_instance_data(obj).set_value(value)

    def __delete__(self, obj):
        pass

    def bind(self, obj, func):
        self._get_instance_data(obj).bind(func)

    def strongbind(self, obj, func):
        self._get_instance_data(obj).strongbind(func)

    def unbind(self, obj, func):
        self._get_instance_data(obj).unbind(func)

    def unbind_all(self, obj):
        self._get_instance_data(obj).unbind_all()

    def __repr__(self):
        return f'ObservableProperty at 0x{id(self):X}, default value={self._default_value}'


def _find_property(obj, name):
    """ Helper that looks in class hierarchy for matching member
    """
    classes = inspect.getmro(
        obj.__class__)  # getmro returns a tuple of class base classes, including class, in method resolution order

    for cls in classes:  # find the first class that has the expected member
        try:
            return cls.__dict__[name]
        except KeyError:
            pass
    raise KeyError(name)  # if no class matches, raise KeyError


def bind(obj, **kwargs):
    """ bind callables with a weak reference.
    Use this bind method for all 'normal' callables like functions or methods.
    The advantage is that the garbage collector can remove objects even if they are referenced by ObservableProperty.
    ObservableProperty silently removes the callable if it no longer exists.
    This method does not work with lambda expressions!
    :param obj: an object with ObservableProperty member(s)
    :param kwargs: name of parameter must match the name of an ObservableProperty, value must be a callable."""
    for name, func in kwargs.items():
        prop = _find_property(obj, name)
        prop.bind(obj, func)


def strongbind(obj, **kwargs):
    """ bind callables with a strong reference.
     This method also works with lambda expressions, but you must unbind the callable before the garbage collector can delete it."""
    for name, func in kwargs.items():
        prop = _find_property(obj, name)
        prop.strongbind(obj, func)


def unbind(obj, **kwargs):
    """ unbind callables that were bound before.
    :param obj: an object with ObservableProperty member(s)
    :param kwargs: name of parameter must match the name of an ObservableProperty, value must be a callable.
                    Unbinding an unknown callable is allowed, in this cases nothing changes. """
    for name, func in kwargs.items():
        prop = _find_property(obj, name)
        prop.unbind(obj, func)


def unbind_all(obj, *propertyNames):
    """ unbind all callables that were bound before.
    :param obj: an object with ObservableProperty member(s)
    :param propertyNames: list of strings, each string names an ObservableProperty.
    """
    for name in propertyNames:
        prop = _find_property(obj, name)
        prop.unbind_all(obj)


@contextmanager
def bound_context(obj, **kwargs):
    """ context manager for bind / unbind sequence."""
    bind(obj, **kwargs)
    try:
        yield
    finally:
        unbind(obj, **kwargs)


@contextmanager
def strong_bound_context(obj, **kwargs):
    """ context manager for strongbind / unbind sequence."""
    strongbind(obj, **kwargs)
    try:
        yield
    finally:
        unbind(obj, **kwargs)



"""Sco Module implements Service Controller Operations(sco) functionality.

All remote control commands of a client are executed by sco instances.

These operations share a common behavior:
A remote control command is executed. The response to such soap request contains a state
(typically 'wait') and a transaction id.
The progress of the transaction is reported with an OperationInvokedReport.
A client must subscribe to the OperationInvokeReport Event of the 'Set' service,
otherwise it would not get informed about progress.
"""
from __future__ import annotations

import queue
import threading
import time
import traceback
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from sdc11073 import loghelper
from sdc11073.exceptions import ApiUsageError

if TYPE_CHECKING:
    from enum import Enum

    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol
    from sdc11073.mdib.providermdib import ProviderMdib
    from sdc11073.pysoap.soapenvelope import ReceivedSoapMessage
    from sdc11073.xml_types.msg_types import AbstractSet
    from sdc11073.roles.providerbase import OperationClassGetter
    from .operations import OperationDefinitionBase
    from .porttypes.setserviceimpl import SetServiceProtocol


class _OperationsWorker(threading.Thread):
    """Thread that enqueues and processes all operations.

    Progress notifications are sent via subscription manager.
    """

    def __init__(self,
                 operations_registry: AbstractScoOperationsRegistry,
                 set_service: SetServiceProtocol,
                 mdib: ProviderMdib,
                 log_prefix: str):
        super().__init__(name='DeviceOperationsWorker')
        self.daemon = True
        self._operations_registry = operations_registry
        self._set_service: SetServiceProtocol = set_service
        self._mdib = mdib
        self._operations_queue = queue.Queue(10)  # spooled operations
        self._logger = loghelper.get_logger_adapter('sdc.device.op_worker', log_prefix)

    def enqueue_operation(self, operation: OperationDefinitionBase,
                          request: ReceivedSoapMessage,
                          operation_request: AbstractSet,
                          transaction_id: int):
        """Enqueue operation."""
        self._operations_queue.put((transaction_id, operation, request, operation_request), timeout=1)

    def run(self):
        data_model = self._mdib.data_model
        InvocationState = data_model.msg_types.InvocationState  # noqa: N806
        InvocationError = data_model.msg_types.InvocationError  # noqa: N806
        while True:
            try:
                try:
                    from_queue = self._operations_queue.get(timeout=1.0)
                except queue.Empty:
                    self._operations_registry.check_invocation_timeouts()
                else:
                    if from_queue == 'stop_sco':
                        self._logger.info('stop request found. Terminating now.')
                        return
                    tr_id, operation, request, operation_request = from_queue  # unpack tuple
                    time.sleep(0.001)
                    self._logger.info('%s: starting operation "%s" argument=%r',
                                      operation.__class__.__name__, operation.handle, operation_request.argument)
                    # duplicate the WAIT response to the operation request as notification. Standard requires this.
                    self._set_service.notify_operation(
                        operation, tr_id, InvocationState.WAIT, self._mdib.mdib_version_group)
                    time.sleep(0.001)  # not really necessary, but in real world there might also be some delay.
                    self._set_service.notify_operation(
                        operation, tr_id, InvocationState.START, self._mdib.mdib_version_group)
                    try:
                        execute_result = operation.execute_operation(request, operation_request)
                        self._logger.info('%s: successfully finished operation "%s"',
                                          operation.__class__.__name__, operation.handle)
                        self._set_service.notify_operation(
                            operation, tr_id, execute_result.invocation_state,
                            self._mdib.mdib_version_group, execute_result.operation_target_handle)
                    except Exception as ex:
                        self._logger.error('%s: error executing operation "%s": %s', operation.__class__.__name__,
                                           operation.handle, traceback.format_exc())
                        self._set_service.notify_operation(
                            operation, tr_id, InvocationState.FAILED, self._mdib.mdib_version_group,
                            error=InvocationError.OTHER, error_message=repr(ex))
            except Exception:
                self._logger.error('%s: unexpected error while handling operation: %s',
                                   self.__class__.__name__, traceback.format_exc())

    def stop(self):
        self._operations_queue.put('stop_sco')  # a dummy request to stop the thread
        self.join(timeout=1)


class AbstractScoOperationsRegistry(ABC):
    """Base class for a Sco."""

    def __init__(self, set_service: SetServiceProtocol,
                 operation_cls_getter: OperationClassGetter,
                 mdib: ProviderMdib,
                 sco_descriptor_container: AbstractDescriptorProtocol,
                 log_prefix: str | None = None):
        self._worker = None
        self._set_service: SetServiceProtocol = set_service
        self.operation_cls_getter = operation_cls_getter
        self._mdib = mdib
        self.sco_descriptor_container = sco_descriptor_container
        self._log_prefix = log_prefix
        self._logger = loghelper.get_logger_adapter('sdc.device.op_reg', log_prefix)
        self._registered_operations = {}  # lookup by handle

    def check_invocation_timeouts(self):
        """Call check_timeout of all registered operations."""
        for op in self._registered_operations.values():
            op.check_timeout()

    @abstractmethod
    def register_operation(self, operation: OperationDefinitionBase) -> None:
        """Register the operation."""

    @abstractmethod
    def unregister_operation_by_handle(self, operation_handle: str) -> None:
        """Un-register the operation."""

    @abstractmethod
    def get_operation_by_handle(self, operation_handle: str) -> OperationDefinitionBase:
        """Get OperationDefinition for given handle."""

    @abstractmethod
    def handle_operation_request(self, operation: OperationDefinitionBase,
                                 request: ReceivedSoapMessage,
                                 operation_request: AbstractSet,
                                 transaction_id: int) -> Enum:
        """Handle operation "operation"."""

    @abstractmethod
    def start_worker(self):
        """Start worker thread."""

    @abstractmethod
    def stop_worker(self):
        """Stop worker thread."""


class ScoOperationsRegistry(AbstractScoOperationsRegistry):
    """Registry for Sco operations.

    from BICEPS:
    A service control object to define remote control operations. Any pm:AbstractOperationDescriptor/@OperationTarget
    within this SCO SHALL only reference this or child descriptors within the CONTAINMENT TREE.
    NOTE - In modular systems, dynamically plugged-in modules would typically be modeled as VMDs.
    Such VMDs potentially have their own SCO. In every other case, SCO operations are modeled in pm:MdsDescriptor/pm:Sco.
    """

    def register_operation(self, operation: OperationDefinitionBase):
        """Register the operation."""
        if operation.handle in self._registered_operations:
            self._logger.debug('handle %s is already registered, will re-use it', operation.handle)
        operation.set_mdib(self._mdib, self.sco_descriptor_container.Handle)
        self._logger.info('register operation "%s"', operation)
        self._registered_operations[operation.handle] = operation

    def unregister_operation_by_handle(self, operation_handle: str):
        """Un-register the operation."""
        del self._registered_operations[operation_handle]

    def get_operation_by_handle(self, operation_handle: str) -> OperationDefinitionBase:
        """Get OperationDefinition for given handle."""
        return self._registered_operations.get(operation_handle)

    def handle_operation_request(self, operation: OperationDefinitionBase,
                                 request: ReceivedSoapMessage,
                                 operation_request: AbstractSet,
                                 transaction_id: int) -> Enum:
        """Handle operation immediately or delayed in worker thread, depending on operation.delayed_processing."""
        InvocationState = self._mdib.data_model.msg_types.InvocationState  # noqa: N806

        if operation.delayed_processing:
            self._worker.enqueue_operation(operation, request, operation_request, transaction_id)
            return InvocationState.WAIT
        try:
            execute_result = operation.execute_operation(request, operation_request)
            self._logger.info('%s: successfully finished operation "%s"', operation.__class__.__name__,
                              operation.handle)
            self._set_service.notify_operation(operation, transaction_id, execute_result.invocation_state,
                                               self._mdib.mdib_version_group, execute_result.operation_target_handle)
            self._logger.debug('notifications for operation %s sent', operation.handle)
            return InvocationState.FINISHED
        except Exception as ex:
            self._logger.error('%s: error executing operation "%s": %s', operation.__class__.__name__,
                               operation.handle, traceback.format_exc())
            self._set_service.notify_operation(
                operation, transaction_id, InvocationState.FAILED, self._mdib.mdib_version_group,
                error=self._mdib.data_model.msg_types.InvocationError.OTHER, error_message=repr(ex))
            return InvocationState.FAILED

    def start_worker(self):
        """Start worker thread."""
        if self._worker is not None:
            raise ApiUsageError('SCO worker is already running')
        self._worker = _OperationsWorker(self, self._set_service, self._mdib, self._log_prefix)
        self._worker.start()

    def stop_worker(self):
        """Stop worker thread."""
        if self._worker is not None:
            self._worker.stop()
            self._worker = None



"""Async implementation of subscriptions manager."""
from __future__ import annotations

import asyncio
import socket
import time
import traceback
from collections import defaultdict
from threading import Thread
from typing import TYPE_CHECKING, Any

import aiohttp.client_exceptions
from lxml import etree

from sdc11073 import observableproperties
from sdc11073.etc import apply_map
from sdc11073.pysoap.soapclient import HTTPReturnCodeError
from sdc11073.xml_types import eventing_types as evt_types
from sdc11073.xml_types.addressing_types import HeaderInformationBlock
from sdc11073.xml_types.basetypes import MessageType

from .subscriptionmgr_base import ActionBasedSubscription, RoundTripData, SubscriptionsManagerBase

if TYPE_CHECKING:
    from collections.abc import Awaitable, Iterable
    from logging import LoggerAdapter

    from sdc11073 import xml_utils
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.dispatch import RequestData
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.pysoap.msgfactory import MessageFactory
    from sdc11073.pysoap.msgreader import ReceivedMessage
    from sdc11073.pysoap.soapclientpool import SoapClientPool


def _mk_dispatch_identifier(reference_parameters: list, path_suffix: str) -> tuple[str | None, str]:
    # this is always our own reference parameter. We know that is has max. one element,
    # and the text is the identifier of the subscription
    if path_suffix == '':
        path_suffix = None
    if len(reference_parameters) > 0:
        return reference_parameters[0].text, path_suffix
    return None, path_suffix


class BicepsSubscriptionAsync(ActionBasedSubscription):
    """Async version of a single BICEPS subscription. It is used by BICEPSSubscriptionsManagerBaseAsync."""

    async def async_send_notification_report(self, body_node: xml_utils.LxmlElement, action: str):
        """Send notification to subscriber."""
        if not self.is_valid or self.unsubscribed_at is not None:
            return
        addr = HeaderInformationBlock(addr_to=self.notify_to_address,
                                      action=action,
                                      addr_from=None,
                                      reference_parameters=self.notify_ref_params)
        message = self._mk_notification_message(addr, body_node)
        try:
            soap_client = self._get_soap_client()
            roundtrip_timer = observableproperties.SingleValueCollector(soap_client, 'roundtrip_time')
            self._logger.debug('send_notification_report {}', action)  # noqa: PLE1205
            await soap_client.async_post_message_to(self.notify_to_url.path, message)
            try:
                roundtrip_time = roundtrip_timer.result(0)
                self.last_roundtrip_times.append(roundtrip_time)
                self.max_roundtrip_time = max(self.max_roundtrip_time, roundtrip_time)
            except observableproperties.CollectTimeoutError:
                pass
            self.notify_errors = 0
            self._is_connection_error = False
        except HTTPReturnCodeError:
            self.notify_errors += 1
            raise
        except asyncio.exceptions.TimeoutError:
            self.notify_errors += 1
            self._is_connection_error = True
            raise
        except Exception:  # any other exception is handled as an unreachable location (disconnected)
            self.notify_errors += 1
            self._is_connection_error = True
            raise

    async def async_send_notification_end_message(
            self,
            code: str = 'SourceShuttingDown',
            reason: str = 'Event source going off line.') -> ReceivedMessage | None:
        """Send notification end message to subscriber."""
        url = self.base_urls[0]
        my_addr = f'{url.scheme}:{url.netloc}/{url.path}'

        if not self.is_valid:
            return None
        soap_client = self._get_soap_client()
        subscription_end = evt_types.SubscriptionEnd()
        subscription_end.SubscriptionManager.Address = my_addr
        subscription_end.SubscriptionManager.ReferenceParameters = self.reference_parameters
        subscription_end.Status = code
        subscription_end.add_reason(reason, 'en-US')
        inf = HeaderInformationBlock(action=subscription_end.action,
                                     addr_to=self.end_to_address or self.notify_to_address,
                                     reference_parameters=self.end_to_ref_params or self.notify_ref_params)
        message = self._msg_factory.mk_soap_message(inf, payload=subscription_end)
        try:
            url = self._end_to_url or self.notify_to_url
            self._logger.info('async send subscription end to {}, subscription = {}', url, self)  # noqa: PLE1205
            return await soap_client.async_post_message_to(url.path, message)
        except aiohttp.client_exceptions.ClientConnectorError:
            # it does not matter that we could not send the message - end is end ;)
            self._logger.info('exception async send subscription end to {}, subscription = {}',  # noqa: PLE1205
                              url, self)
        except Exception:  # noqa: BLE001
            self._logger.error(traceback.format_exc()) # noqa: TRY400
        finally:
            self._is_closed = True


class AsyncioEventLoopThread(Thread):
    """Central event loop for provider."""

    def __init__(self, name: str, logger: LoggerAdapter ):
        super().__init__(name=name)
        self._logger = logger
        self.daemon = True
        self.loop = asyncio.new_event_loop()
        self._running = False

    @property
    def running(self) -> bool:
        """Return True if the event loop is running."""
        return self._running

    def run(self):
        """Run method of thread."""
        self._logger.info('%s started', self.__class__.__name__)
        self._running = True
        self.loop.run_forever()
        self._logger.info('%s finished', self.__class__.__name__)

    def run_coro(self, coro: Awaitable) -> Any:
        """Run threadsafe."""
        if not self._running:
            self._logger.error('%s: async thread is not running', self.__class__.__name__)
            return None
        return asyncio.run_coroutine_threadsafe(coro, loop=self.loop).result()

    def stop(self):
        """Stop thread."""
        self._logger.info('%s: stopping now', self.__class__.__name__)
        self._running = False
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.join()
        self._logger.info('%s: stopped', self.__class__.__name__)


class BICEPSSubscriptionsManagerBaseAsync(SubscriptionsManagerBase):
    """Async version of a subscriptions manager.

    It sends a notification parallel to all subscribers by using the async functionality.
    First all notifications are sent, then all http responses are collected.
    This saves a lot of time compared to the synchronous version that sends the notification to the first subscriber,
    waits for the response, then sends the notification to the second subscriber, and so on.
    """

    def __init__(self,
                 sdc_definitions: BaseDefinitions,
                 msg_factory: MessageFactory,
                 soap_client_pool: SoapClientPool,
                 max_subscription_duration: float | None = None,
                 log_prefix: str | None = None,
                 ):
        super().__init__(sdc_definitions, msg_factory, soap_client_pool, max_subscription_duration, log_prefix)
        if soap_client_pool.async_loop_subscr_mgr is None:
            thr = AsyncioEventLoopThread(name='async_loop_subscr_mgr', logger=self._logger)
            soap_client_pool.async_loop_subscr_mgr = thr
            thr.start()
            for _i in range(10):
                if not thr.running:
                    time.sleep(0.1)
                else:
                    break
            if not thr.running:
                raise RuntimeError('could not start AsyncioEventLoopThread')
        self._async_send_thread = soap_client_pool.async_loop_subscr_mgr

    def _mk_subscription_instance(self, request_data: RequestData) -> BicepsSubscriptionAsync:
        subscribe_request = evt_types.Subscribe.from_node(request_data.message_data.p_msg.msg_node)
        accepted_encodings = request_data.http_header['Accept-Encoding']
        return BicepsSubscriptionAsync(self, subscribe_request, accepted_encodings, self.base_urls,
                                       self._max_subscription_duration,
                                       self._soap_client_pool,
                                       msg_factory=self._msg_factory,
                                       log_prefix=self._logger.log_prefix)

    async def _coro_send_to_subscribers(self, tasks: Iterable[Awaitable]) -> tuple[BaseException | Any, ...]:
        return await asyncio.gather(*tasks, return_exceptions=True)

    def _end_all_subscriptions(self, send_subscription_end: bool):
        self._logger.info('end all subscriptions')
        tasks = []
        with self._subscriptions.lock:
            all_subscriptions = list(self._subscriptions.objects)
            if send_subscription_end:
                for subscription in all_subscriptions:
                    if subscription.is_valid and subscription.unsubscribed_at is None:
                        self._logger.info('send subscription end, subscription = {}', subscription)  # noqa: PLE1205
                        tasks.append(subscription.async_send_notification_end_message())
            if tasks:
                result = self._async_send_thread.run_coro(self._coro_send_to_subscribers(tasks))
                for counter, element in enumerate(result):
                    if isinstance(element, Exception):
                        self._logger.warning(  # noqa: PLE1205
                            'end_all_subscriptions {} returned {}', all_subscriptions[counter], element)

        apply_map(lambda subscription: subscription.close_by_subscription_manager(), self._subscriptions.objects)
        self._subscriptions.clear()

    def send_to_subscribers(self, payload: MessageType | xml_utils.LxmlElement,
                            action: str,
                            mdib_version_group: MdibVersionGroup):
        """Send payload to all subscribers."""
        with self._subscriptions.lock:
            if not self._async_send_thread.running:
                self._logger.warning('could not send notifications, async send loop is not running.')
                self._logger.warning(traceback.format_stack())
                return
            subscribers = self._get_subscriptions_for_action(action)
            if isinstance(payload, MessageType):
                nsh = self.sdc_definitions.data_model.ns_helper
                namespaces = [nsh.PM, nsh.MSG]
                namespaces.extend(payload.additional_namespaces)
                my_ns_map = nsh.partial_map(*namespaces)
                body_node = payload.as_etree_node(payload.NODETYPE, my_ns_map)
            else:
                body_node = payload

            self.sent_to_subscribers = (action, mdib_version_group, body_node)  # update observable
            tasks = []
            for subscriber in subscribers:
                tasks.append(self._async_send_notification_report(subscriber, body_node, action)) # noqa: PERF401

            self._logger.debug('sending report %s to %r',
                               action, [s.notify_to_address for s in subscribers])
            result = self._async_send_thread.run_coro(self._coro_send_to_subscribers(tasks))
            if result is None:
                self._logger.info('could not send notifications, async send loop is not running.')
                return
            for counter, element in enumerate(result):
                if isinstance(element, Exception):
                    self._logger.warning(  # noqa: PLE1205
                        '{}: _send_to_subscribers {} returned {}', action, subscribers[counter], element)

    async def _async_send_notification_report(self, subscription: BicepsSubscriptionAsync,
                                              body_node: xml_utils.LxmlElement,
                                              action: str):
        try:
            self._logger.debug('send notification report {} to {}', action, subscription)  # noqa: PLE1205
            await subscription.async_send_notification_report(body_node, action)
            self._logger.debug(' done: send notification report {} to {}', action, subscription)  # noqa: PLE1205
        except HTTPReturnCodeError as ex:
            # this is an error related to the connection => log warning and continue
            self._logger.warning(  # noqa: PLE1205
                'could not send notification report {}: HTTP status= {}, reason={}, {}',
                action, ex.status, ex.reason, subscription)
        except (aiohttp.client_exceptions.ClientConnectionError,
                aiohttp.client_exceptions.ClientConnectorError,
                aiohttp.client_exceptions.ServerConnectionError,
                asyncio.exceptions.TimeoutError,
                socket.timeout) as ex:
            # this is an error related to the connection => log warning and continue
            self._logger.warning('could not send notification report {} warning= {!r}: {}',  # noqa: PLE1205
                                 action, ex, subscription)
        except etree.DocumentInvalid as ex:
            # this is an error related to the document, it cannot be sent to any subscriber => re-raise
            self._logger.warning('Invalid Document for action {}: {!r}\n{}',  # noqa: PLE1205
                                 action, ex, etree.tostring(body_node))
            raise
        except Exception:
            # this should never happen! => re-raise
            self._logger.error('could not send notification report {}, error {}: {}',  # noqa: PLE1205 TRY400
                               action, traceback.format_exc(), subscription)
            raise

    def get_subscription_round_trip_times(self) -> dict[tuple[str, tuple[str]], RoundTripData]:
        """Calculate round trip times based on last MAX_ROUNDTRIP_VALUES values.

        :return: a dictionary with key=(<notify_to_address>, (subscription_names)),
                value = RoundTripData with members min, max, avg, abs_max, values
        """
        ret = {}
        with self._subscriptions.lock:
            for subscription in self._subscriptions.objects:
                if subscription.max_roundtrip_time > 0:
                    ret[(subscription.notify_to_address,
                         subscription.short_filter_names())] = subscription.get_roundtrip_stats()
        return ret

    def get_client_round_trip_times(self) -> dict[str, RoundTripData]:
        """Calculate round trip times based on last MAX_ROUNDTRIP_VALUES values.

        :return: a dictionary with key=<notify_to_address>,
        value = _RoundTripData with members min, max, avg, abs_max, values
        """
        # first step: collect all round trip times of subscriptions, group them by notify_to_address
        tmp = defaultdict(list)
        ret = {}
        with self._subscriptions.lock:
            for subscription in self._subscriptions.objects:
                if subscription.max_roundtrip_time > 0:
                    tmp[subscription.notify_to_address].append(subscription.get_roundtrip_stats())
        for key, stats in tmp.items():
            all_values = [stat.values for stat in stats]
            ret[key] = RoundTripData(all_values, max([s.max for s in stats]))
        return ret


class SubscriptionsManagerPathAsync(BICEPSSubscriptionsManagerBaseAsync):
    """Use path dispatching to identify subscriptions."""

    def _mk_subscription_instance(self, request_data: RequestData) -> BicepsSubscriptionAsync:
        subscription = super()._mk_subscription_instance(request_data)
        subscription.path_suffix = subscription.identifier_uuid.hex
        return subscription


class SubscriptionsManagerReferenceParamAsync(BICEPSSubscriptionsManagerBaseAsync):
    """Use reference parameters to identify subscriptions."""

    def _mk_subscription_instance(self, request_data: RequestData) -> BicepsSubscriptionAsync:
        subscription = super()._mk_subscription_instance(request_data)
        # add  a reference parameter
        subscription.set_reference_parameter()
        return subscription



from __future__ import annotations

import time
from collections import deque
from concurrent import futures
from dataclasses import dataclass
from enum import IntEnum
from statistics import mean, stdev
from threading import Lock
from typing import TYPE_CHECKING, Callable, ClassVar

from sdc11073 import observableproperties as properties
from sdc11073.exceptions import ApiUsageError

if TYPE_CHECKING:
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.pysoap.msgreader import ReceivedMessage

    from .consumermdib import ConsumerMdib
    from .statecontainers import AbstractMetricStateContainer, AbstractStateProtocol




PROFILING = False
if PROFILING:
    import cProfile
    import pstats
    from io import StringIO

class _WarningState(IntEnum):
    A_NO_LOG = 0
    A_OUT_OF_RANGE = 1
    A_STILL_OUT_OF_RANGE = 2
    A_BACK_IN_RANGE = 3

LOG_WF_AGE_INTERVAL = 30  # how often a log message is written with mean and stdev of waveforms age
AGE_CALC_SAMPLES_COUNT = 100  # amount of data for wf mean age and stdev calculation


class DeterminationTimeWarner:
    """A Helper to reduce log warnings regarding determination time."""

    ST_IN_RANGE = 0
    ST_OUT_OF_RANGE = 1
    result_lookup: ClassVar[dict[tuple[int, int], tuple[_WarningState, bool]]] = {
        (ST_IN_RANGE, ST_IN_RANGE): (_WarningState.A_NO_LOG, False),
        (ST_IN_RANGE, ST_OUT_OF_RANGE): (_WarningState.A_OUT_OF_RANGE, False),
        (ST_OUT_OF_RANGE, ST_OUT_OF_RANGE): (_WarningState.A_STILL_OUT_OF_RANGE, True),
        (ST_OUT_OF_RANGE, ST_IN_RANGE): (_WarningState.A_BACK_IN_RANGE, False),
    }

    def __init__(self, repeat_period: int=30):
        """Construct the DeterminationTimeWarner.

        :param repeat_period: period after which an existing warning condition the warning shall be repeated.
        """
        self.repeat_period = repeat_period
        self._last_log_time = 0
        self.last_state = self.ST_IN_RANGE

    def get_out_of_determination_time_log_state(self, min_age: float,
                                                max_age: float,
                                                warn_limit: float) -> _WarningState:
        """:return: one of above constants."""
        now = time.time()
        current_state = self.ST_OUT_OF_RANGE if min_age < -warn_limit or max_age > warn_limit else self.ST_IN_RANGE
        warning_state, shall_repeat = self.result_lookup[(self.last_state, current_state)]
        if self.last_state != current_state:
            # a state transition
            self.last_state = current_state
            self._last_log_time = now
            return warning_state
        # no state transition, but might need repeated logging
        if shall_repeat and now - self._last_log_time >= self.repeat_period:
            self._last_log_time = now
            return warning_state
        return _WarningState.A_NO_LOG


class AgeLogger:
    """A helper that write to lo  if incoming states are too old or in the future."""

    def __init__(self, metric_time_warner: DeterminationTimeWarner,
                 warn_limit: float,
                 log_prefix: str,
                 mdib_version: int):
        self._metric_time_warner = metric_time_warner
        self._warn_limit = warn_limit
        self._log_prefix = log_prefix
        self._mdib_version = mdib_version
        self.age_list: list[float] = []
        self.now = time.time()

    def add_determination_time(self, determination_time: float):
        """Make the determination time an age and add it to self.age_list."""
        self.age_list.append(self.now - determination_time)

    def log_age_warnings(self, logger: LoggerAdapter):
        """Write age warning to logger."""
        if len(self.age_list) == 0:
            return
        min_age = min(self.age_list)
        max_age = max(self.age_list)

        shall_log = self._metric_time_warner.get_out_of_determination_time_log_state(
            min_age, max_age, self._warn_limit)
        if shall_log == _WarningState.A_OUT_OF_RANGE:
            logger.warning(  # noqa: PLE1205
                '{} mdib_version {}: age of states outside limit of {} sec.: max, min = {:03f}, {:03f}',
                self._log_prefix, self._mdib_version, self._warn_limit, max_age, min_age)
        elif shall_log == _WarningState.A_STILL_OUT_OF_RANGE:
            logger.warning( # noqa: PLE1205
                '{} mdib_version {}: age of states still outside limit of {} sec.: max, min = {:03f}, {:03f}',
                self._log_prefix, self._mdib_version, self._warn_limit, max_age, min_age)
        elif shall_log == _WarningState.A_BACK_IN_RANGE:
            logger.info( # noqa: PLE1205
                '{} mdib_version {}: age of states back in limit of {} sec.: max, min = {:03f}, {:03f}',
                self._log_prefix, self._mdib_version, self._warn_limit, max_age, min_age)


@dataclass
class AgeData:
    """Container for some statistical age data."""

    mean_age: float
    stdev: float
    min_age: float
    max_age: float


class AgeStatistics:
    """Keep age data of a single state over time."""

    min_list_lenght_for_statistics = 3
    def __init__(self, entry_count: int | None = None):
        length = entry_count or AGE_CALC_SAMPLES_COUNT
        self._age_of_data_list = deque(
            maxlen=length)  # used to calculate average age of samples when received
        self._lock = Lock()

    def process_state(self, metric_state: AbstractMetricStateContainer):
        """Append age of metric_state to internal list."""
        try:
            age = time.time() - metric_state.MetricValue.DeterminationTime
            with self._lock:
                self._age_of_data_list.append(age)
        except AttributeError as ex:
            if not metric_state.is_metric_state:
                raise ApiUsageError(f'{self.__class__.__name__} can only handle metric states') from ex
            # or state.MetricValue is None or  DeterminationTime is None: ignore this

    def get_age_stdev(self) -> AgeData:
        """Calculate age statistics data."""
        if len(self._age_of_data_list) < self.min_list_lenght_for_statistics:
            return AgeData(0.0, 0.0, 0.0, 0.0)
        with self._lock:
            min_value = min(self._age_of_data_list)
            max_value = max(self._age_of_data_list)
            mean_data = mean(self._age_of_data_list)
            std_deviation = stdev(self._age_of_data_list)
            return AgeData(mean_data, std_deviation, min_value, max_value)


class ConsumerMdibMethods:
    """Extra methods for consumer mdib tht are not core functionality."""

    DETERMINATIONTIME_WARN_LIMIT = 1.0  # in seconds

    def __init__(self, consumer_mdib: ConsumerMdib, logger: LoggerAdapter):
        self._mdib = consumer_mdib
        self._sdc_client = consumer_mdib.sdc_client
        self._msg_reader = self._sdc_client.msg_reader
        self._logger = logger
        self.waveform_time_warner = DeterminationTimeWarner()
        self.metric_time_warner = DeterminationTimeWarner()
        self._last_wf_age_log = time.time()
        if PROFILING:
            self.prof = cProfile.Profile()
        self._age_statistics = {}
        self._calculate_wf_age_stats = False

    def set_calculate_wf_age_stats(self, shall_calculate: bool):
        """Switch calculation of statistice on or off."""
        self._calculate_wf_age_stats = shall_calculate

    def wait_metric_matches(self, handle: str,
                            matches_func: Callable[[AbstractMetricStateContainer], bool],
                            timeout: float) -> AbstractMetricStateContainer:
        """Wait until a matching metric has been received.

        The matching is defined by the handle of the metric
        and the result of a matching function. If the matching function returns true, this function returns.
        :param handle: The handle string of the metric of interest.
        :param matches_func: a callable, argument is the current state with matching handle.
                              Can be None, in that case every state matches
        Example:
            expected = 42
            def isMatchingValue(state):
                if state.MetricValue is None:
                    return False.

                found_value = state.MetricValue.Value
                return [expected] == found_value
        :param timeout: timeout in seconds
        :return: the matching state. In case of a timeout it raises a TimeoutError exception.
        """
        fut = futures.Future()

        # define a callback function that sets value of fut
        def on_metrics_by_handle(metrics_by_handle: dict[str, AbstractMetricStateContainer]):
            metric = metrics_by_handle.get(handle)
            if metric is not None:
                if matches_func is None or matches_func(metric):
                    fut.set_result(metric)

        try:
            properties.bind(self._mdib, metrics_by_handle=on_metrics_by_handle)
            begin = time.monotonic()
            ret = fut.result(timeout)
            self._logger.debug('wait_metric_matches: got result after {:.2f} seconds',  # noqa: PLE1205
                               time.monotonic() - begin)
            return ret
        finally:
            properties.unbind(self._mdib, metrics_by_handle=on_metrics_by_handle)

    def mk_proposed_state(self, descriptor_handle: str,
                          copy_current_state: bool = True,
                          handle: str | None = None) -> AbstractStateProtocol:
        """Create a new state that can be used as proposed state in according operations.

        The new state is not part of mdib!.

        :param descriptor_handle: the descriptor
        :param copy_current_state: if True, all members of existing state will be copied to new state
        :param handle: if this is a multi state class, and the handle is not None,
                      this is the handle of the existing state that shall be used for copy.
        :return: a new state container
        """
        descr = self._mdib.descriptions.handle.get_one(descriptor_handle)
        new_state = self._mdib.data_model.mk_state_container(descr)
        if copy_current_state:
            lookup = self._mdib.context_states if new_state.is_context_state else self._mdib.states
            if new_state.is_multi_state:
                if handle is None:  # new state
                    return new_state
                old_state = lookup.handle.get_one(handle)
            else:
                old_state = lookup.descriptor_handle.get_one(descriptor_handle)
            new_state.update_from_other_container(old_state)
        return new_state

    def sync_context_states(self):
        """Sync context states with provider.

        Requests all context states from device and deletes all local context states that are not
        available in response from Device.
        """
        self._logger.info('_sync_context_states called')
        context_service = self._sdc_client.client('Context')
        response = context_service.get_context_states()
        context_state_containers = response.result.ContextState

        devices_context_state_handles = [s.Handle for s in context_state_containers]
        with self._mdib.context_states.lock:
            for obj in self._mdib.context_states.objects:
                if obj.Handle not in devices_context_state_handles:
                    self._mdib.context_states.remove_object_no_lock(obj)

    def bind_to_client_observables(self):
        """Connect the mdib with the notifications from consumer."""
        if PROFILING:
            properties.bind(self._sdc_client, waveform_report=self._on_waveform_report_profiled)
        else:
            properties.bind(self._sdc_client, waveform_report=self._on_waveform_report)
        properties.bind(self._sdc_client, episodic_metric_report=self._on_episodic_metric_report)
        properties.bind(self._sdc_client, episodic_alert_report=self._on_episodic_alert_report)
        properties.bind(self._sdc_client, episodic_context_report=self._on_episodic_context_report)
        properties.bind(self._sdc_client, episodic_component_report=self._on_episodic_component_report)
        properties.bind(self._sdc_client, description_modification_report=self._on_description_modification_report)
        properties.bind(self._sdc_client, episodic_operational_state_report=self._on_operational_state_report)

    def _on_episodic_metric_report(self, received_message_data: ReceivedMessage):
        model = self._mdib.data_model
        cls = model.msg_types.EpisodicMetricReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_metric_states_report(received_message_data.mdib_version_group, report)

        if not self._mdib.is_initialized:
            return

        # generate warnings if age of states is out of accepted range
        age_logger = AgeLogger(self.metric_time_warner, self.DETERMINATIONTIME_WARN_LIMIT,
                               'EpisodicMetricReport', self._mdib.mdib_version)
        for report_part in report.ReportPart:
            for state_container in report_part.values_list:
                desc_h = state_container.DescriptorHandle
                if state_container.MetricValue is not None:
                    # BICEPS: While Validity is "Ong" or "NA", the enclosing METRIC value SHALL not possess a
                    # determined value.
                    # Also ignore determination time if measurement is invalid or not active.
                    if state_container.ActivationState == model.pm_types.ComponentActivation.ON and \
                            state_container.MetricValue.MetricQuality.Validity not in [
                        model.pm_types.MeasurementValidity.INVALID,
                        model.pm_types.MeasurementValidity.NA,
                        model.pm_types.MeasurementValidity.MEASUREMENT_ONGOING]:
                        determination_time = state_container.MetricValue.DeterminationTime
                        if determination_time is None:
                            self._logger.warning(  # noqa: PLE1205
                                'EpisodicMetricReport: metric {} version {} has no DeterminationTime',
                                desc_h, state_container.StateVersion)
                        else:
                            age_logger.add_determination_time(determination_time)

        age_logger.log_age_warnings(self._logger)

    def _on_episodic_alert_report(self, received_message_data: ReceivedMessage):
        cls = self._mdib.data_model.msg_types.EpisodicAlertReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_alert_states_report(received_message_data.mdib_version_group, report)

    def _on_operational_state_report(self, received_message_data: ReceivedMessage):
        cls = self._mdib.data_model.msg_types.EpisodicOperationalStateReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_operational_states_report(received_message_data.mdib_version_group, report)

    def _on_waveform_report_profiled(self, received_message_data: ReceivedMessage):
        """Handle waveform report and print profiling information to stdout."""
        reference_count_print_limit = 50
        self.prof.enable()
        self._on_waveform_report(received_message_data)
        self.prof.disable()
        str_io = StringIO()
        stats = pstats.Stats(self.prof, stream=str_io).sort_stats('cumulative')
        stats.print_stats(30)
        print(str_io.getvalue())
        print(f'total number of states: {len(self._mdib.states.objects)}')
        print(f'total number of objIds: {len(self._mdib.states._object_ids)}')  # noqa: SLF001
        for name, refs in self._mdib.states._object_ids.items():  # noqa: SLF001
            if len(refs) > reference_count_print_limit:
                print(f'object {name} has {len(refs)} idx references, {refs}')

    def _on_waveform_report(self, received_message_data: ReceivedMessage):
        """Handle waveform report."""
        cls = self._mdib.data_model.msg_types.WaveformStream
        report = cls.from_node(received_message_data.p_msg.msg_node)
        if self._calculate_wf_age_stats:
            self._process_wf_age_statistics(report.State)
        accepted_states = self._mdib.process_incoming_waveform_states(received_message_data.mdib_version_group,
                                                                      report.State)

        if accepted_states is None or len(accepted_states) == 0 or not self._mdib.is_initialized:
            return

        waveform_age = {}  # collect age of all waveforms in this report, and make one report if age is above warn limit (instead of multiple)
        now = time.time()
        for state_container in accepted_states.values():
            rt_sample_containers = self._mdib.rt_buffers[state_container.DescriptorHandle].rt_data
            if len(rt_sample_containers) > 0:
                waveform_age[state_container.DescriptorHandle] = now - rt_sample_containers[-1].determination_time

        if len(waveform_age) > 0:
            min_age = min(waveform_age.values())
            max_age = max(waveform_age.values())
            shall_log = self.waveform_time_warner.get_out_of_determination_time_log_state(
                min_age, max_age, self.DETERMINATIONTIME_WARN_LIMIT)
            if shall_log != _WarningState.A_NO_LOG:
                tmp = ', '.join(f'"{k}": {v:.3f}sec.' for k, v in waveform_age.items())
                if shall_log == _WarningState.A_OUT_OF_RANGE:
                    self._logger.warning(  # noqa: PLE1205
                        '_on_waveform_report mdib_version {}: age of samples outside limit of {} sec.: {}',
                        self._mdib.mdib_version, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
                elif shall_log == _WarningState.A_STILL_OUT_OF_RANGE:
                    self._logger.warning(  # noqa: PLE1205
                        '_on_waveform_report mdib_version {}: age of samples still outside limit of {} sec.: {}',
                        self._mdib.mdib_version, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
                elif shall_log == _WarningState.A_BACK_IN_RANGE:
                    self._logger.info(  # noqa: PLE1205
                        '_on_waveform_report mdib_version {}: age of samples back in limit of {} sec.: {}',
                        self._mdib.mdib_version, self.DETERMINATIONTIME_WARN_LIMIT, tmp)
        if LOG_WF_AGE_INTERVAL:
            now = time.time()
            if now - self._last_wf_age_log >= LOG_WF_AGE_INTERVAL:
                age_data = self.get_wf_age_stdev()
                if age_data is not None:
                    self._logger.info(  # noqa: PLE1205
                        'waveform mean age={:.1f}ms., stdev={:.2f}ms. min={:.1f}ms., max={}',
                        age_data.mean_age * 1000., age_data.stdev * 1000.,
                        age_data.min_age * 1000., age_data.max_age * 1000.)
                self._last_wf_age_log = now

    def _on_episodic_context_report(self, received_message_data: ReceivedMessage):
        """Handle episodic context report."""
        cls = self._mdib.data_model.msg_types.EpisodicContextReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_context_states_report(received_message_data.mdib_version_group, report)

    def _on_episodic_component_report(self, received_message_data: ReceivedMessage):
        """Handle episodic component report.

        The EpisodicComponentReport is sent if at least one property of at least one component state has changed
        and SHOULD contain only the changed component states.
        Components are MDSs, VMDs, Channels. Not metrics and alarms.
        """
        cls = self._mdib.data_model.msg_types.EpisodicComponentReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_component_states_report(received_message_data.mdib_version_group, report)

    def _on_description_modification_report(self, received_message_data: ReceivedMessage):
        """Handle description modification report.

        The EpisodicComponentReport is sent if at least one property of at least one component state has changed
        and SHOULD contain only the changed component states.
        Components are MDSs, VMDs, Channels. Not metrics and alarms.
        """
        cls = self._mdib.data_model.msg_types.DescriptionModificationReport
        report = cls.from_node(received_message_data.p_msg.msg_node)
        self._mdib.process_incoming_description_modifications(received_message_data.mdib_version_group, report)

    def _process_wf_age_statistics(self, state_containers: list[AbstractMetricStateContainer]):
        for st in state_containers:
            age_stat = self._age_statistics.get(st.DescriptorHandle)
            if age_stat is None:
                age_stat = AgeStatistics()
                self._age_statistics[st.DescriptorHandle] = age_stat
            age_stat.process_state(st)

    def get_wf_age_stdev(self) -> AgeData:
        """Create some statistics data for age of waveform data when it arrived.

        Data is used for logging.
        """
        if len(self._age_statistics) < AgeStatistics.min_list_lenght_for_statistics:
            return AgeData(0.0, 0.0, 0.0, 0.0)

        means = []
        stdevs = []
        mins = []
        maxs = []
        for age_stat in self._age_statistics.values():
            age_data = age_stat.get_age_stdev()
            means.append(age_data.mean_age)
            stdevs.append(age_data.stdev)
            mins.append(age_data.min_age)
            maxs.append(age_data.max_age)
        return AgeData(mean(means), mean(stdevs), min(mins), max(maxs))
